"""AUTOGENERATED. DO NOT EDIT."""

from tf_keras.src.applications.resnet import ResNet50
from tf_keras.src.applications.resnet import decode_predictions
from tf_keras.src.applications.resnet import preprocess_input



# Copyright 2020 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Benchmark for TF-Keras text vectorization preprocessing layer's adapt method.
"""

import collections
import itertools
import random
import string
import time

import numpy as np
import tensorflow.compat.v2 as tf

import tf_keras.src as keras
from tf_keras.src.layers.preprocessing import index_lookup

tf.compat.v1.enable_v2_behavior()


# word_gen creates random sequences of ASCII letters (both lowercase and upper).
# The number of unique strings is ~2,700.
def word_gen():
    for _ in itertools.count(1):
        yield "".join(random.choice(string.ascii_letters) for i in range(2))


def get_top_k(dataset, k):
    """Python implementation of vocabulary building using a defaultdict."""
    counts = collections.defaultdict(int)
    for tensor in dataset:
        data = tensor.numpy()
        for element in data:
            counts[element] += 1
    sorted_vocab = [
        k
        for k, _ in sorted(
            counts.items(), key=lambda item: item[1], reverse=True
        )
    ]
    if len(sorted_vocab) > k:
        sorted_vocab = sorted_vocab[:k]
    return sorted_vocab


class BenchmarkAdapt(tf.test.Benchmark):
    """Benchmark adapt."""

    def run_numpy_implementation(self, num_elements, batch_size, k):
        """Test the python implementation."""
        ds = tf.data.Dataset.from_generator(
            word_gen, tf.string, tf.TensorShape([])
        )
        batched_ds = ds.take(num_elements).batch(batch_size)
        input_t = keras.Input(shape=(), dtype=tf.string)
        layer = index_lookup.IndexLookup(
            max_tokens=k,
            num_oov_indices=0,
            mask_token=None,
            oov_token="OOV",
            dtype=tf.string,
        )
        _ = layer(input_t)
        num_repeats = 5
        starts = []
        ends = []
        for _ in range(num_repeats):
            starts.append(time.time())
            vocab = get_top_k(batched_ds, k)
            layer.set_vocabulary(vocab)
            ends.append(time.time())
        avg_time = np.mean(np.array(ends) - np.array(starts))
        return avg_time

    def bm_adapt_implementation(self, num_elements, batch_size, k):
        """Test the KPL adapt implementation."""
        ds = tf.data.Dataset.from_generator(
            word_gen, tf.string, tf.TensorShape([])
        )
        batched_ds = ds.take(num_elements).batch(batch_size)
        input_t = keras.Input(shape=(), dtype=tf.string)
        layer = index_lookup.IndexLookup(
            max_tokens=k,
            num_oov_indices=0,
            mask_token=None,
            oov_token="OOV",
            dtype=tf.string,
        )
        _ = layer(input_t)
        num_repeats = 5
        starts = []
        ends = []
        for _ in range(num_repeats):
            starts.append(time.time())
            layer.adapt(batched_ds)
            ends.append(time.time())
        avg_time = np.mean(np.array(ends) - np.array(starts))
        name = "index_lookup_adapt|%s_elements|vocab_size_%s|batch_%s" % (
            num_elements,
            k,
            batch_size,
        )
        baseline = self.run_numpy_implementation(num_elements, batch_size, k)
        extras = {
            "numpy implementation baseline": baseline,
            "delta seconds": (baseline - avg_time),
            "delta percent": ((baseline - avg_time) / baseline) * 100,
        }
        self.report_benchmark(
            iters=num_repeats, wall_time=avg_time, extras=extras, name=name
        )

    def benchmark_vocab_size_by_batch(self):
        for vocab_size in [100, 1000, 10000, 100000, 1000000]:
            for batch in [1, 16, 2048]:
                self.bm_adapt_implementation(
                    vocab_size, batch, int(vocab_size / 10)
                )


if __name__ == "__main__":
    tf.test.main()




"""AUTOGENERATED. DO NOT EDIT."""

from tf_keras.api._v1.keras.preprocessing import image
from tf_keras.api._v1.keras.preprocessing import sequence
from tf_keras.api._v1.keras.preprocessing import text



"""AUTOGENERATED. DO NOT EDIT."""

from tf_keras.src.applications.resnet_rs import ResNetRS101
from tf_keras.src.applications.resnet_rs import ResNetRS152
from tf_keras.src.applications.resnet_rs import ResNetRS200
from tf_keras.src.applications.resnet_rs import ResNetRS270
from tf_keras.src.applications.resnet_rs import ResNetRS350
from tf_keras.src.applications.resnet_rs import ResNetRS420
from tf_keras.src.applications.resnet_rs import ResNetRS50
from tf_keras.src.applications.resnet_rs import decode_predictions
from tf_keras.src.applications.resnet_rs import preprocess_input



"""AUTOGENERATED. DO NOT EDIT."""

from tf_keras.src.preprocessing.image import DirectoryIterator
from tf_keras.src.preprocessing.image import ImageDataGenerator
from tf_keras.src.preprocessing.image import Iterator
from tf_keras.src.preprocessing.image import NumpyArrayIterator
from tf_keras.src.preprocessing.image import apply_affine_transform
from tf_keras.src.preprocessing.image import apply_brightness_shift
from tf_keras.src.preprocessing.image import apply_channel_shift
from tf_keras.src.preprocessing.image import random_brightness
from tf_keras.src.preprocessing.image import random_channel_shift
from tf_keras.src.preprocessing.image import random_rotation
from tf_keras.src.preprocessing.image import random_shear
from tf_keras.src.preprocessing.image import random_shift
from tf_keras.src.preprocessing.image import random_zoom
from tf_keras.src.utils.image_utils import array_to_img
from tf_keras.src.utils.image_utils import img_to_array
from tf_keras.src.utils.image_utils import load_img
from tf_keras.src.utils.image_utils import save_img



"""AUTOGENERATED. DO NOT EDIT."""

from tf_keras.src.preprocessing.sequence import TimeseriesGenerator
from tf_keras.src.preprocessing.sequence import make_sampling_table
from tf_keras.src.preprocessing.sequence import skipgrams
from tf_keras.src.utils.data_utils import pad_sequences



# Copyright 2015 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Contains the SpatialDropout3D layer."""


import tensorflow.compat.v2 as tf

from tf_keras.src import backend
from tf_keras.src.engine.input_spec import InputSpec
from tf_keras.src.layers.regularization.dropout import Dropout

# isort: off
from tensorflow.python.util.tf_export import keras_export


@keras_export("keras.layers.SpatialDropout3D")
class SpatialDropout3D(Dropout):
    """Spatial 3D version of Dropout.

    This version performs the same function as Dropout, however, it drops
    entire 3D feature maps instead of individual elements. If adjacent voxels
    within feature maps are strongly correlated (as is normally the case in
    early convolution layers) then regular dropout will not regularize the
    activations and will otherwise just result in an effective learning rate
    decrease. In this case, SpatialDropout3D will help promote independence
    between feature maps and should be used instead.

    Args:
      rate: Float between 0 and 1. Fraction of the input units to drop.
      data_format: 'channels_first' or 'channels_last'. In 'channels_first'
        mode, the channels dimension (the depth) is at index 1, in
        'channels_last' mode is it at index 4. When unspecified, uses
        `image_data_format` value found in your TF-Keras config file at
        `~/.keras/keras.json` (if exists) else 'channels_last'.
        Defaults to 'channels_last'.
    Call arguments:
      inputs: A 5D tensor.
      training: Python boolean indicating whether the layer should behave in
        training mode (adding dropout) or in inference mode (doing nothing).
    Input shape:
      5D tensor with shape: `(samples, channels, dim1, dim2, dim3)` if
        data_format='channels_first'
      or 5D tensor with shape: `(samples, dim1, dim2, dim3, channels)` if
        data_format='channels_last'.
    Output shape: Same as input.
    References: - [Efficient Object Localization Using Convolutional
        Networks](https://arxiv.org/abs/1411.4280)
    """

    def __init__(self, rate, data_format=None, **kwargs):
        super().__init__(rate, **kwargs)
        if data_format is None:
            data_format = backend.image_data_format()
        if data_format not in {"channels_last", "channels_first"}:
            raise ValueError(
                '`data_format` must be "channels_last" or "channels_first". '
                f"Received: data_format={data_format}."
            )
        self.data_format = data_format
        self.input_spec = InputSpec(ndim=5)

    def _get_noise_shape(self, inputs):
        input_shape = tf.shape(inputs)
        if self.data_format == "channels_first":
            return (input_shape[0], input_shape[1], 1, 1, 1)
        elif self.data_format == "channels_last":
            return (input_shape[0], 1, 1, 1, input_shape[4])




"""AUTOGENERATED. DO NOT EDIT."""

from tf_keras.src.applications.mobilenet import MobileNet
from tf_keras.src.applications.mobilenet import decode_predictions
from tf_keras.src.applications.mobilenet import preprocess_input



"""AUTOGENERATED. DO NOT EDIT."""

from tf_keras.src.initializers import deserialize
from tf_keras.src.initializers import get
from tf_keras.src.initializers import serialize
from tf_keras.src.initializers.initializers import Constant
from tf_keras.src.initializers.initializers import Constant as constant
from tf_keras.src.initializers.initializers import GlorotNormal
from tf_keras.src.initializers.initializers import GlorotNormal as glorot_normal
from tf_keras.src.initializers.initializers import GlorotUniform
from tf_keras.src.initializers.initializers import GlorotUniform as glorot_uniform
from tf_keras.src.initializers.initializers import HeNormal
from tf_keras.src.initializers.initializers import HeNormal as he_normal
from tf_keras.src.initializers.initializers import HeUniform
from tf_keras.src.initializers.initializers import HeUniform as he_uniform
from tf_keras.src.initializers.initializers import Identity
from tf_keras.src.initializers.initializers import Identity as identity
from tf_keras.src.initializers.initializers import Initializer
from tf_keras.src.initializers.initializers import LecunNormal
from tf_keras.src.initializers.initializers import LecunNormal as lecun_normal
from tf_keras.src.initializers.initializers import LecunUniform
from tf_keras.src.initializers.initializers import LecunUniform as lecun_uniform
from tf_keras.src.initializers.initializers import Ones
from tf_keras.src.initializers.initializers import Ones as ones
from tf_keras.src.initializers.initializers import Orthogonal
from tf_keras.src.initializers.initializers import Orthogonal as orthogonal
from tf_keras.src.initializers.initializers import RandomNormal
from tf_keras.src.initializers.initializers import RandomNormal as random_normal
from tf_keras.src.initializers.initializers import RandomUniform
from tf_keras.src.initializers.initializers import RandomUniform as random_uniform
from tf_keras.src.initializers.initializers import TruncatedNormal
from tf_keras.src.initializers.initializers import TruncatedNormal as truncated_normal
from tf_keras.src.initializers.initializers import VarianceScaling
from tf_keras.src.initializers.initializers import VarianceScaling as variance_scaling
from tf_keras.src.initializers.initializers import Zeros
from tf_keras.src.initializers.initializers import Zeros as zeros



"""AUTOGENERATED. DO NOT EDIT."""

from tf_keras.src.activations import deserialize
from tf_keras.src.activations import elu
from tf_keras.src.activations import exponential
from tf_keras.src.activations import gelu
from tf_keras.src.activations import get
from tf_keras.src.activations import hard_sigmoid
from tf_keras.src.activations import linear
from tf_keras.src.activations import mish
from tf_keras.src.activations import relu
from tf_keras.src.activations import selu
from tf_keras.src.activations import serialize
from tf_keras.src.activations import sigmoid
from tf_keras.src.activations import softmax
from tf_keras.src.activations import softplus
from tf_keras.src.activations import softsign
from tf_keras.src.activations import swish
from tf_keras.src.activations import tanh



"""AUTOGENERATED. DO NOT EDIT."""

from tf_keras.api._v1.keras.datasets import boston_housing
from tf_keras.api._v1.keras.datasets import cifar10
from tf_keras.api._v1.keras.datasets import cifar100
from tf_keras.api._v1.keras.datasets import fashion_mnist
from tf_keras.api._v1.keras.datasets import imdb
from tf_keras.api._v1.keras.datasets import mnist
from tf_keras.api._v1.keras.datasets import reuters



"""AUTOGENERATED. DO NOT EDIT."""

from tf_keras.src.constraints import Constraint
from tf_keras.src.constraints import MaxNorm
from tf_keras.src.constraints import MaxNorm as max_norm
from tf_keras.src.constraints import MinMaxNorm
from tf_keras.src.constraints import MinMaxNorm as min_max_norm
from tf_keras.src.constraints import NonNeg
from tf_keras.src.constraints import NonNeg as non_neg
from tf_keras.src.constraints import RadialConstraint
from tf_keras.src.constraints import RadialConstraint as radial_constraint
from tf_keras.src.constraints import UnitNorm
from tf_keras.src.constraints import UnitNorm as unit_norm
from tf_keras.src.constraints import deserialize
from tf_keras.src.constraints import get
from tf_keras.src.constraints import serialize



# Copyright 2016 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Keras layers API."""

# isort: off
import tensorflow.compat.v2 as tf

from tf_keras.src.engine.base_layer import Layer
from tf_keras.src.engine.base_preprocessing_layer import PreprocessingLayer

# Generic layers.
from tf_keras.src.engine.input_layer import Input
from tf_keras.src.engine.input_layer import InputLayer
from tf_keras.src.engine.input_spec import InputSpec
from tf_keras.src.layers.activation.elu import ELU
from tf_keras.src.layers.activation.leaky_relu import LeakyReLU
from tf_keras.src.layers.activation.prelu import PReLU

# Activations layers.
from tf_keras.src.layers.activation.relu import ReLU
from tf_keras.src.layers.activation.softmax import Softmax
from tf_keras.src.layers.activation.thresholded_relu import ThresholdedReLU
from tf_keras.src.layers.attention.additive_attention import AdditiveAttention
from tf_keras.src.layers.attention.attention import Attention

# Attention layers.
from tf_keras.src.layers.attention.multi_head_attention import MultiHeadAttention

# Convolution layer aliases.
# Convolution layers.
from tf_keras.src.layers.convolutional.conv1d import Conv1D
from tf_keras.src.layers.convolutional.conv1d import Convolution1D
from tf_keras.src.layers.convolutional.conv1d_transpose import Conv1DTranspose
from tf_keras.src.layers.convolutional.conv1d_transpose import (
    Convolution1DTranspose,
)
from tf_keras.src.layers.convolutional.conv2d import Conv2D
from tf_keras.src.layers.convolutional.conv2d import Convolution2D
from tf_keras.src.layers.convolutional.conv2d_transpose import Conv2DTranspose
from tf_keras.src.layers.convolutional.conv2d_transpose import (
    Convolution2DTranspose,
)
from tf_keras.src.layers.convolutional.conv3d import Conv3D
from tf_keras.src.layers.convolutional.conv3d import Convolution3D
from tf_keras.src.layers.convolutional.conv3d_transpose import Conv3DTranspose
from tf_keras.src.layers.convolutional.conv3d_transpose import (
    Convolution3DTranspose,
)
from tf_keras.src.layers.convolutional.depthwise_conv1d import DepthwiseConv1D
from tf_keras.src.layers.convolutional.depthwise_conv2d import DepthwiseConv2D
from tf_keras.src.layers.convolutional.separable_conv1d import SeparableConv1D
from tf_keras.src.layers.convolutional.separable_conv1d import (
    SeparableConvolution1D,
)
from tf_keras.src.layers.convolutional.separable_conv2d import SeparableConv2D
from tf_keras.src.layers.convolutional.separable_conv2d import (
    SeparableConvolution2D,
)

# Core layers.
from tf_keras.src.layers.core.activation import Activation
from tf_keras.src.layers.core.dense import Dense
from tf_keras.src.layers.core.einsum_dense import EinsumDense
from tf_keras.src.layers.core.embedding import Embedding
from tf_keras.src.layers.core.identity import Identity
from tf_keras.src.layers.core.lambda_layer import Lambda
from tf_keras.src.layers.core.masking import Masking
from tf_keras.src.layers.core.tf_op_layer import ClassMethod
from tf_keras.src.layers.core.tf_op_layer import InstanceMethod
from tf_keras.src.layers.core.tf_op_layer import InstanceProperty
from tf_keras.src.layers.core.tf_op_layer import SlicingOpLambda
from tf_keras.src.layers.core.tf_op_layer import TFOpLambda

# Locally-connected layers.
from tf_keras.src.layers.locally_connected.locally_connected1d import (
    LocallyConnected1D,
)
from tf_keras.src.layers.locally_connected.locally_connected2d import (
    LocallyConnected2D,
)

# Merging functions.
# Merging layers.
from tf_keras.src.layers.merging.add import Add
from tf_keras.src.layers.merging.add import add
from tf_keras.src.layers.merging.average import Average
from tf_keras.src.layers.merging.average import average
from tf_keras.src.layers.merging.concatenate import Concatenate
from tf_keras.src.layers.merging.concatenate import concatenate
from tf_keras.src.layers.merging.dot import Dot
from tf_keras.src.layers.merging.dot import dot
from tf_keras.src.layers.merging.maximum import Maximum
from tf_keras.src.layers.merging.maximum import maximum
from tf_keras.src.layers.merging.minimum import Minimum
from tf_keras.src.layers.merging.minimum import minimum
from tf_keras.src.layers.merging.multiply import Multiply
from tf_keras.src.layers.merging.multiply import multiply
from tf_keras.src.layers.merging.subtract import Subtract
from tf_keras.src.layers.merging.subtract import subtract
from tf_keras.src.layers.normalization.batch_normalization import (
    SyncBatchNormalization,
)

# Normalization layers.
from tf_keras.src.layers.normalization.group_normalization import GroupNormalization
from tf_keras.src.layers.normalization.layer_normalization import LayerNormalization
from tf_keras.src.layers.normalization.unit_normalization import UnitNormalization
from tf_keras.src.layers.normalization.spectral_normalization import (
    SpectralNormalization,
)  # noqa: E501

# Preprocessing layers.
from tf_keras.src.layers.preprocessing.category_encoding import CategoryEncoding
from tf_keras.src.layers.preprocessing.discretization import Discretization
from tf_keras.src.layers.preprocessing.hashed_crossing import HashedCrossing
from tf_keras.src.layers.preprocessing.hashing import Hashing

# Image preprocessing layers.
from tf_keras.src.layers.preprocessing.image_preprocessing import CenterCrop
from tf_keras.src.layers.preprocessing.image_preprocessing import RandomBrightness
from tf_keras.src.layers.preprocessing.image_preprocessing import RandomContrast
from tf_keras.src.layers.preprocessing.image_preprocessing import RandomCrop
from tf_keras.src.layers.preprocessing.image_preprocessing import RandomFlip
from tf_keras.src.layers.preprocessing.image_preprocessing import RandomHeight
from tf_keras.src.layers.preprocessing.image_preprocessing import RandomRotation
from tf_keras.src.layers.preprocessing.image_preprocessing import RandomTranslation
from tf_keras.src.layers.preprocessing.image_preprocessing import RandomWidth
from tf_keras.src.layers.preprocessing.image_preprocessing import RandomZoom
from tf_keras.src.layers.preprocessing.image_preprocessing import Rescaling
from tf_keras.src.layers.preprocessing.image_preprocessing import Resizing
from tf_keras.src.layers.preprocessing.integer_lookup import IntegerLookup
from tf_keras.src.layers.preprocessing.normalization import Normalization
from tf_keras.src.layers.preprocessing.string_lookup import StringLookup
from tf_keras.src.layers.preprocessing.text_vectorization import TextVectorization
from tf_keras.src.layers.regularization.activity_regularization import (
    ActivityRegularization,
)
from tf_keras.src.layers.regularization.alpha_dropout import AlphaDropout

# Regularization layers.
from tf_keras.src.layers.regularization.dropout import Dropout
from tf_keras.src.layers.regularization.gaussian_dropout import GaussianDropout
from tf_keras.src.layers.regularization.gaussian_noise import GaussianNoise
from tf_keras.src.layers.regularization.spatial_dropout1d import SpatialDropout1D
from tf_keras.src.layers.regularization.spatial_dropout2d import SpatialDropout2D
from tf_keras.src.layers.regularization.spatial_dropout3d import SpatialDropout3D

# Reshaping layers.
from tf_keras.src.layers.reshaping.cropping1d import Cropping1D
from tf_keras.src.layers.reshaping.cropping2d import Cropping2D
from tf_keras.src.layers.reshaping.cropping3d import Cropping3D
from tf_keras.src.layers.reshaping.flatten import Flatten
from tf_keras.src.layers.reshaping.permute import Permute
from tf_keras.src.layers.reshaping.repeat_vector import RepeatVector
from tf_keras.src.layers.reshaping.reshape import Reshape
from tf_keras.src.layers.reshaping.up_sampling1d import UpSampling1D
from tf_keras.src.layers.reshaping.up_sampling2d import UpSampling2D
from tf_keras.src.layers.reshaping.up_sampling3d import UpSampling3D
from tf_keras.src.layers.reshaping.zero_padding1d import ZeroPadding1D
from tf_keras.src.layers.reshaping.zero_padding2d import ZeroPadding2D
from tf_keras.src.layers.reshaping.zero_padding3d import ZeroPadding3D

if tf.__internal__.tf2.enabled():
    from tf_keras.src.layers.normalization.batch_normalization import (
        BatchNormalization,
    )
    from tf_keras.src.layers.normalization.batch_normalization_v1 import (
        BatchNormalization as BatchNormalizationV1,
    )

    BatchNormalizationV2 = BatchNormalization
else:
    from tf_keras.src.layers.normalization.batch_normalization import (
        BatchNormalization as BatchNormalizationV2,
    )
    from tf_keras.src.layers.normalization.batch_normalization_v1 import (
        BatchNormalization,
    )

    BatchNormalizationV1 = BatchNormalization

# Kernelized layers.
from tf_keras.src.layers.kernelized import RandomFourierFeatures

# Pooling layer aliases.
# Pooling layers.
from tf_keras.src.layers.pooling.average_pooling1d import AveragePooling1D
from tf_keras.src.layers.pooling.average_pooling1d import AvgPool1D
from tf_keras.src.layers.pooling.average_pooling2d import AveragePooling2D
from tf_keras.src.layers.pooling.average_pooling2d import AvgPool2D
from tf_keras.src.layers.pooling.average_pooling3d import AveragePooling3D
from tf_keras.src.layers.pooling.average_pooling3d import AvgPool3D
from tf_keras.src.layers.pooling.global_average_pooling1d import (
    GlobalAveragePooling1D,
)
from tf_keras.src.layers.pooling.global_average_pooling1d import GlobalAvgPool1D
from tf_keras.src.layers.pooling.global_average_pooling2d import (
    GlobalAveragePooling2D,
)
from tf_keras.src.layers.pooling.global_average_pooling2d import GlobalAvgPool2D
from tf_keras.src.layers.pooling.global_average_pooling3d import (
    GlobalAveragePooling3D,
)
from tf_keras.src.layers.pooling.global_average_pooling3d import GlobalAvgPool3D
from tf_keras.src.layers.pooling.global_max_pooling1d import GlobalMaxPool1D
from tf_keras.src.layers.pooling.global_max_pooling1d import GlobalMaxPooling1D
from tf_keras.src.layers.pooling.global_max_pooling2d import GlobalMaxPool2D
from tf_keras.src.layers.pooling.global_max_pooling2d import GlobalMaxPooling2D
from tf_keras.src.layers.pooling.global_max_pooling3d import GlobalMaxPool3D
from tf_keras.src.layers.pooling.global_max_pooling3d import GlobalMaxPooling3D
from tf_keras.src.layers.pooling.max_pooling1d import MaxPool1D
from tf_keras.src.layers.pooling.max_pooling1d import MaxPooling1D
from tf_keras.src.layers.pooling.max_pooling2d import MaxPool2D
from tf_keras.src.layers.pooling.max_pooling2d import MaxPooling2D
from tf_keras.src.layers.pooling.max_pooling3d import MaxPool3D
from tf_keras.src.layers.pooling.max_pooling3d import MaxPooling3D
from tf_keras.src.layers.rnn.abstract_rnn_cell import AbstractRNNCell

# Recurrent layers.
from tf_keras.src.layers.rnn.base_rnn import RNN
from tf_keras.src.layers.rnn.simple_rnn import SimpleRNN
from tf_keras.src.layers.rnn.simple_rnn import SimpleRNNCell
from tf_keras.src.layers.rnn.stacked_rnn_cells import StackedRNNCells

if tf.__internal__.tf2.enabled():
    from tf_keras.src.layers.rnn.gru import GRU
    from tf_keras.src.layers.rnn.gru import GRUCell
    from tf_keras.src.layers.rnn.gru_v1 import GRU as GRUV1
    from tf_keras.src.layers.rnn.gru_v1 import GRUCell as GRUCellV1
    from tf_keras.src.layers.rnn.lstm import LSTM
    from tf_keras.src.layers.rnn.lstm import LSTMCell
    from tf_keras.src.layers.rnn.lstm_v1 import LSTM as LSTMV1
    from tf_keras.src.layers.rnn.lstm_v1 import LSTMCell as LSTMCellV1

    GRUV2 = GRU
    GRUCellV2 = GRUCell
    LSTMV2 = LSTM
    LSTMCellV2 = LSTMCell
else:
    from tf_keras.src.layers.rnn.gru import GRU as GRUV2
    from tf_keras.src.layers.rnn.gru import GRUCell as GRUCellV2
    from tf_keras.src.layers.rnn.gru_v1 import GRU
    from tf_keras.src.layers.rnn.gru_v1 import GRUCell
    from tf_keras.src.layers.rnn.lstm import LSTM as LSTMV2
    from tf_keras.src.layers.rnn.lstm import LSTMCell as LSTMCellV2
    from tf_keras.src.layers.rnn.lstm_v1 import LSTM
    from tf_keras.src.layers.rnn.lstm_v1 import LSTMCell

    GRUV1 = GRU
    GRUCellV1 = GRUCell
    LSTMV1 = LSTM
    LSTMCellV1 = LSTMCell

# Serialization functions.
from tf_keras.src.layers import serialization

# Wrapper functions.
from tf_keras.src.layers.rnn.base_wrapper import Wrapper
from tf_keras.src.layers.rnn.bidirectional import Bidirectional

# RNN Cell wrappers.
from tf_keras.src.layers.rnn.cell_wrappers import DeviceWrapper
from tf_keras.src.layers.rnn.cell_wrappers import DropoutWrapper
from tf_keras.src.layers.rnn.cell_wrappers import ResidualWrapper

# Convolutional-recurrent layers.
from tf_keras.src.layers.rnn.conv_lstm1d import ConvLSTM1D
from tf_keras.src.layers.rnn.conv_lstm2d import ConvLSTM2D
from tf_keras.src.layers.rnn.conv_lstm3d import ConvLSTM3D
from tf_keras.src.layers.rnn.cudnn_gru import CuDNNGRU

# cuDNN recurrent layers.
from tf_keras.src.layers.rnn.cudnn_lstm import CuDNNLSTM
from tf_keras.src.layers.rnn.time_distributed import TimeDistributed
from tf_keras.src.layers.serialization import deserialize
from tf_keras.src.layers.serialization import deserialize_from_json
from tf_keras.src.layers.serialization import get_builtin_layer
from tf_keras.src.layers.serialization import serialize


class VersionAwareLayers:
    """Utility to be used internally to access layers in a V1/V2-aware fashion.

    When using layers within the TF-Keras codebase, under the constraint that
    e.g. `layers.BatchNormalization` should be the `BatchNormalization` version
    corresponding to the current runtime (TF1 or TF2), do not simply access
    `layers.BatchNormalization` since it would ignore e.g. an early
    `compat.v2.disable_v2_behavior()` call. Instead, use an instance
    of `VersionAwareLayers` (which you can use just like the `layers` module).
    """

    def __getattr__(self, name):
        serialization.populate_deserializable_objects()
        if name in serialization.LOCAL.ALL_OBJECTS:
            return serialization.LOCAL.ALL_OBJECTS[name]
        return super().__getattr__(name)




# Copyright 2015 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Python utilities required by TF-Keras."""

import inspect
import threading

# isort: off
from tensorflow.python.util.tf_export import keras_export

_GLOBAL_CUSTOM_OBJECTS = {}
_GLOBAL_CUSTOM_NAMES = {}
# Thread-local custom objects set by custom_object_scope.
_THREAD_LOCAL_CUSTOM_OBJECTS = threading.local()


@keras_export(
    "keras.saving.custom_object_scope",
    "keras.utils.custom_object_scope",
    "keras.utils.CustomObjectScope",
)
class CustomObjectScope:
    """Exposes custom classes/functions to TF-Keras deserialization internals.

    Under a scope `with custom_object_scope(objects_dict)`, TF-Keras methods
    such as `tf.keras.models.load_model` or `tf.keras.models.model_from_config`
    will be able to deserialize any custom object referenced by a
    saved config (e.g. a custom layer or metric).

    Example:

    Consider a custom regularizer `my_regularizer`:

    ```python
    layer = Dense(3, kernel_regularizer=my_regularizer)
    # Config contains a reference to `my_regularizer`
    config = layer.get_config()
    ...
    # Later:
    with custom_object_scope({'my_regularizer': my_regularizer}):
      layer = Dense.from_config(config)
    ```

    Args:
        *args: Dictionary or dictionaries of `{name: object}` pairs.
    """

    def __init__(self, *args):
        self.custom_objects = args
        self.backup = None

    def __enter__(self):
        self.backup = _THREAD_LOCAL_CUSTOM_OBJECTS.__dict__.copy()
        for objects in self.custom_objects:
            _THREAD_LOCAL_CUSTOM_OBJECTS.__dict__.update(objects)
        return self

    def __exit__(self, *args, **kwargs):
        _THREAD_LOCAL_CUSTOM_OBJECTS.__dict__.clear()
        _THREAD_LOCAL_CUSTOM_OBJECTS.__dict__.update(self.backup)


@keras_export(
    "keras.saving.get_custom_objects", "keras.utils.get_custom_objects"
)
def get_custom_objects():
    """Retrieves a live reference to the global dictionary of custom objects.

    Custom objects set using using `custom_object_scope` are not added to the
    global dictionary of custom objects, and will not appear in the returned
    dictionary.

    Example:

    ```python
    get_custom_objects().clear()
    get_custom_objects()['MyObject'] = MyObject
    ```

    Returns:
        Global dictionary mapping registered class names to classes.
    """
    return _GLOBAL_CUSTOM_OBJECTS


@keras_export(
    "keras.saving.register_keras_serializable",
    "keras.utils.register_keras_serializable",
)
def register_keras_serializable(package="Custom", name=None):
    """Registers an object with the TF-Keras serialization framework.

    This decorator injects the decorated class or function into the TF-Keras
    custom object dictionary, so that it can be serialized and deserialized
    without needing an entry in the user-provided custom object dict. It also
    injects a function that TF-Keras will call to get the object's serializable
    string key.

    Note that to be serialized and deserialized, classes must implement the
    `get_config()` method. Functions do not have this requirement.

    The object will be registered under the key 'package>name' where `name`,
    defaults to the object name if not passed.

    Example:

    ```python
    # Note that `'my_package'` is used as the `package` argument here, and since
    # the `name` argument is not provided, `'MyDense'` is used as the `name`.
    @keras.saving.register_keras_serializable('my_package')
    class MyDense(keras.layers.Dense):
      pass

    assert keras.saving.get_registered_object('my_package>MyDense') == MyDense
    assert keras.saving.get_registered_name(MyDense) == 'my_package>MyDense'
    ```

    Args:
      package: The package that this class belongs to. This is used for the
        `key` (which is `"package>name"`) to idenfify the class. Note that this
        is the first argument passed into the decorator.
      name: The name to serialize this class under in this package. If not
        provided or `None`, the class' name will be used (note that this is the
        case when the decorator is used with only one argument, which becomes
        the `package`).

    Returns:
      A decorator that registers the decorated class with the passed names.
    """

    def decorator(arg):
        """Registers a class with the TF-Keras serialization framework."""
        class_name = name if name is not None else arg.__name__
        registered_name = package + ">" + class_name

        if inspect.isclass(arg) and not hasattr(arg, "get_config"):
            raise ValueError(
                "Cannot register a class that does not have a "
                "get_config() method."
            )

        _GLOBAL_CUSTOM_OBJECTS[registered_name] = arg
        _GLOBAL_CUSTOM_NAMES[arg] = registered_name

        return arg

    return decorator


@keras_export(
    "keras.saving.get_registered_name", "keras.utils.get_registered_name"
)
def get_registered_name(obj):
    """Returns the name registered to an object within the TF-Keras framework.

    This function is part of the TF-Keras serialization and deserialization
    framework. It maps objects to the string names associated with those objects
    for serialization/deserialization.

    Args:
      obj: The object to look up.

    Returns:
      The name associated with the object, or the default Python name if the
        object is not registered.
    """
    if obj in _GLOBAL_CUSTOM_NAMES:
        return _GLOBAL_CUSTOM_NAMES[obj]
    else:
        return obj.__name__


@keras_export(
    "keras.saving.get_registered_object", "keras.utils.get_registered_object"
)
def get_registered_object(name, custom_objects=None, module_objects=None):
    """Returns the class associated with `name` if it is registered with
    TF-Keras.

    This function is part of the TF-Keras serialization and deserialization
    framework. It maps strings to the objects associated with them for
    serialization/deserialization.

    Example:

    ```python
    def from_config(cls, config, custom_objects=None):
      if 'my_custom_object_name' in config:
        config['hidden_cls'] = tf.keras.saving.get_registered_object(
            config['my_custom_object_name'], custom_objects=custom_objects)
    ```

    Args:
      name: The name to look up.
      custom_objects: A dictionary of custom objects to look the name up in.
        Generally, custom_objects is provided by the user.
      module_objects: A dictionary of custom objects to look the name up in.
        Generally, module_objects is provided by midlevel library implementers.

    Returns:
      An instantiable class associated with `name`, or `None` if no such class
        exists.
    """
    if name in _THREAD_LOCAL_CUSTOM_OBJECTS.__dict__:
        return _THREAD_LOCAL_CUSTOM_OBJECTS.__dict__[name]
    elif name in _GLOBAL_CUSTOM_OBJECTS:
        return _GLOBAL_CUSTOM_OBJECTS[name]
    elif custom_objects and name in custom_objects:
        return custom_objects[name]
    elif module_objects and name in module_objects:
        return module_objects[name]
    return None


# Aliases
custom_object_scope = CustomObjectScope




"""AUTOGENERATED. DO NOT EDIT."""

from tf_keras.src.datasets.reuters import get_label_names
from tf_keras.src.datasets.reuters import get_word_index
from tf_keras.src.datasets.reuters import load_data



"""AUTOGENERATED. DO NOT EDIT."""

from tf_keras.src.applications.mobilenet_v2 import MobileNetV2
from tf_keras.src.applications.mobilenet_v2 import decode_predictions
from tf_keras.src.applications.mobilenet_v2 import preprocess_input



"""AUTOGENERATED. DO NOT EDIT."""

from tf_keras.src.layers.rnn.legacy_cell_wrappers import DeviceWrapper
from tf_keras.src.layers.rnn.legacy_cell_wrappers import DropoutWrapper
from tf_keras.src.layers.rnn.legacy_cell_wrappers import ResidualWrapper
from tf_keras.src.layers.rnn.legacy_cells import BasicLSTMCell
from tf_keras.src.layers.rnn.legacy_cells import BasicRNNCell
from tf_keras.src.layers.rnn.legacy_cells import GRUCell
from tf_keras.src.layers.rnn.legacy_cells import LSTMCell
from tf_keras.src.layers.rnn.legacy_cells import LSTMStateTuple
from tf_keras.src.layers.rnn.legacy_cells import MultiRNNCell
from tf_keras.src.layers.rnn.legacy_cells import RNNCell



"""AUTOGENERATED. DO NOT EDIT."""

from tf_keras.api._v1.keras.__internal__.legacy.layers import experimental
from tf_keras.src.engine.input_spec import InputSpec
from tf_keras.src.legacy_tf_layers.base import Layer
from tf_keras.src.legacy_tf_layers.convolutional import Conv1D
from tf_keras.src.legacy_tf_layers.convolutional import Conv2D
from tf_keras.src.legacy_tf_layers.convolutional import Conv2DTranspose
from tf_keras.src.legacy_tf_layers.convolutional import Conv3D
from tf_keras.src.legacy_tf_layers.convolutional import Conv3DTranspose
from tf_keras.src.legacy_tf_layers.convolutional import SeparableConv1D
from tf_keras.src.legacy_tf_layers.convolutional import SeparableConv2D
from tf_keras.src.legacy_tf_layers.convolutional import conv1d
from tf_keras.src.legacy_tf_layers.convolutional import conv2d
from tf_keras.src.legacy_tf_layers.convolutional import conv2d_transpose
from tf_keras.src.legacy_tf_layers.convolutional import conv3d
from tf_keras.src.legacy_tf_layers.convolutional import conv3d_transpose
from tf_keras.src.legacy_tf_layers.convolutional import separable_conv1d
from tf_keras.src.legacy_tf_layers.convolutional import separable_conv2d
from tf_keras.src.legacy_tf_layers.core import Dense
from tf_keras.src.legacy_tf_layers.core import Dropout
from tf_keras.src.legacy_tf_layers.core import Flatten
from tf_keras.src.legacy_tf_layers.core import dense
from tf_keras.src.legacy_tf_layers.core import dropout
from tf_keras.src.legacy_tf_layers.core import flatten
from tf_keras.src.legacy_tf_layers.normalization import BatchNormalization
from tf_keras.src.legacy_tf_layers.normalization import batch_normalization
from tf_keras.src.legacy_tf_layers.pooling import AveragePooling1D
from tf_keras.src.legacy_tf_layers.pooling import AveragePooling2D
from tf_keras.src.legacy_tf_layers.pooling import AveragePooling3D
from tf_keras.src.legacy_tf_layers.pooling import MaxPooling1D
from tf_keras.src.legacy_tf_layers.pooling import MaxPooling2D
from tf_keras.src.legacy_tf_layers.pooling import MaxPooling3D
from tf_keras.src.legacy_tf_layers.pooling import average_pooling1d
from tf_keras.src.legacy_tf_layers.pooling import average_pooling2d
from tf_keras.src.legacy_tf_layers.pooling import average_pooling3d
from tf_keras.src.legacy_tf_layers.pooling import max_pooling1d
from tf_keras.src.legacy_tf_layers.pooling import max_pooling2d
from tf_keras.src.legacy_tf_layers.pooling import max_pooling3d



"""AUTOGENERATED. DO NOT EDIT."""

from tf_keras.src.losses import BinaryCrossentropy
from tf_keras.src.losses import BinaryFocalCrossentropy
from tf_keras.src.losses import CategoricalCrossentropy
from tf_keras.src.losses import CategoricalFocalCrossentropy
from tf_keras.src.losses import CategoricalHinge
from tf_keras.src.losses import CosineSimilarity
from tf_keras.src.losses import Hinge
from tf_keras.src.losses import Huber
from tf_keras.src.losses import KLDivergence
from tf_keras.src.losses import LogCosh
from tf_keras.src.losses import Loss
from tf_keras.src.losses import MeanAbsoluteError
from tf_keras.src.losses import MeanAbsolutePercentageError
from tf_keras.src.losses import MeanSquaredError
from tf_keras.src.losses import MeanSquaredLogarithmicError
from tf_keras.src.losses import Poisson
from tf_keras.src.losses import SparseCategoricalCrossentropy
from tf_keras.src.losses import SquaredHinge
from tf_keras.src.losses import binary_crossentropy
from tf_keras.src.losses import binary_focal_crossentropy
from tf_keras.src.losses import categorical_crossentropy
from tf_keras.src.losses import categorical_focal_crossentropy
from tf_keras.src.losses import categorical_hinge
from tf_keras.src.losses import cosine_similarity
from tf_keras.src.losses import deserialize
from tf_keras.src.losses import get
from tf_keras.src.losses import hinge
from tf_keras.src.losses import huber
from tf_keras.src.losses import kl_divergence
from tf_keras.src.losses import kl_divergence as KLD
from tf_keras.src.losses import kl_divergence as kld
from tf_keras.src.losses import kl_divergence as kullback_leibler_divergence
from tf_keras.src.losses import log_cosh
from tf_keras.src.losses import log_cosh as logcosh
from tf_keras.src.losses import mean_absolute_error
from tf_keras.src.losses import mean_absolute_error as MAE
from tf_keras.src.losses import mean_absolute_error as mae
from tf_keras.src.losses import mean_absolute_percentage_error
from tf_keras.src.losses import mean_absolute_percentage_error as MAPE
from tf_keras.src.losses import mean_absolute_percentage_error as mape
from tf_keras.src.losses import mean_squared_error
from tf_keras.src.losses import mean_squared_error as MSE
from tf_keras.src.losses import mean_squared_error as mse
from tf_keras.src.losses import mean_squared_logarithmic_error
from tf_keras.src.losses import mean_squared_logarithmic_error as MSLE
from tf_keras.src.losses import mean_squared_logarithmic_error as msle
from tf_keras.src.losses import poisson
from tf_keras.src.losses import serialize
from tf_keras.src.losses import sparse_categorical_crossentropy
from tf_keras.src.losses import squared_hinge
from tf_keras.src.utils.losses_utils import ReductionV2 as Reduction



"""AUTOGENERATED. DO NOT EDIT."""

from tf_keras.api._v1.keras.layers.experimental import preprocessing
from tf_keras.src.layers.core.einsum_dense import EinsumDense
from tf_keras.src.layers.kernelized import RandomFourierFeatures



"""AUTOGENERATED. DO NOT EDIT."""

from tf_keras.api._v1.keras import __internal__
from tf_keras.api._v1.keras import activations
from tf_keras.api._v1.keras import applications
from tf_keras.api._v1.keras import backend
from tf_keras.api._v1.keras import callbacks
from tf_keras.api._v1.keras import constraints
from tf_keras.api._v1.keras import datasets
from tf_keras.api._v1.keras import estimator
from tf_keras.api._v1.keras import experimental
from tf_keras.api._v1.keras import export
from tf_keras.api._v1.keras import initializers
from tf_keras.api._v1.keras import layers
from tf_keras.api._v1.keras import losses
from tf_keras.api._v1.keras import metrics
from tf_keras.api._v1.keras import mixed_precision
from tf_keras.api._v1.keras import models
from tf_keras.api._v1.keras import optimizers
from tf_keras.api._v1.keras import preprocessing
from tf_keras.api._v1.keras import regularizers
from tf_keras.api._v1.keras import saving
from tf_keras.api._v1.keras import utils
from tf_keras.src.engine.input_layer import Input
from tf_keras.src.engine.sequential import Sequential
from tf_keras.src.engine.training import Model



"""AUTOGENERATED. DO NOT EDIT."""

from tf_keras.src.applications.efficientnet import EfficientNetB0
from tf_keras.src.applications.efficientnet import EfficientNetB1
from tf_keras.src.applications.efficientnet import EfficientNetB2
from tf_keras.src.applications.efficientnet import EfficientNetB3
from tf_keras.src.applications.efficientnet import EfficientNetB4
from tf_keras.src.applications.efficientnet import EfficientNetB5
from tf_keras.src.applications.efficientnet import EfficientNetB6
from tf_keras.src.applications.efficientnet import EfficientNetB7
from tf_keras.src.applications.efficientnet import decode_predictions
from tf_keras.src.applications.efficientnet import preprocess_input



# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Contains AutoCastVariable, a variable which automatically casts itself."""

import threading
from typing import Optional

import tensorflow.compat.v2 as tf

from tf_keras.src.distribute import distributed_training_utils

# _autocast_dtype.dtype is the dtype AutoCastVariables should be cast to, or
# None if AutoCastVariables should not be cast.
_autocast_dtype = threading.local()


def numpy_text(tensor, is_repr=False):
    """Human readable representation of a tensor's numpy value."""
    if tensor.dtype.is_numpy_compatible:
        text = repr(tensor._numpy()) if is_repr else str(tensor._numpy())

    else:
        text = "<unprintable>"
    if "\n" in text:
        text = "\n" + text
    return text


class AutoCastVariableSpec(tf.types.experimental.TraceType):
    """TraceType for AutoCastVariableSpec for tracing with tf.function.

    This class implements the Type for AutoCastVariable used in tracing.
    """

    def __init__(self, value):
        self._value = value

    def is_subtype_of(self, other) -> bool:
        """If the other spec is the same as `self`, return True."""
        return self == other

    def most_specific_common_supertype(self, others):
        """`self` is the common supertype if all input types match it."""
        return self if all(self == other for other in others) else None

    def placeholder_value(self, placeholder_context=None):
        """Use the AutoCastVariable value itself as a placeholder."""
        return self._value

    def cast(self, value, _):
        return value

    def to_tensors(self, value):
        return []

    def __hash__(self) -> int:
        return hash(id(self._value))

    def __eq__(self, other) -> bool:
        return self is other


class AutoCastVariable(tf.Variable, tf.__internal__.types.Tensor):
    """Variable that casts itself to a different dtype in applicable contexts.

    This class wraps a floating-point `tf.Variable`. It emulates the variable
    interface and delegates to the wrapped variable, but it additionally will
    cast the wrapped variable under an `enable_auto_cast_variables(dtype)`
    context manager.

    For example:

    >>> v = tf.Variable(1.0, dtype=tf.float32)
    >>> v = AutoCastVariable(v)
    >>> tf.identity(v).dtype
    tf.float32
    >>> with enable_auto_cast_variables(tf.float16):
    ...   tf.identity(v).dtype
    tf.float16

    The purpose of this class is to allow TF-Keras layers to create variables in
    float32, and automatically cast them to float16 or bfloat16 when the layer
    is called.
    """

    def __init__(self, variable):
        """Creates an AutoCastVariable instance.

        Args:
          variable: A floating-point resource variable to wrap.

        Raises:
          ValueError: If `variable` is not a floating-point resource variable
        """
        if not isinstance(variable, tf.Variable):
            raise ValueError(
                "variable must be of type tf.ResourceVariable, but got: %s"
                % variable
            )
        if not variable.dtype.is_floating:
            raise ValueError(
                "variable must be a floating point variable but has type: %s"
                % variable.dtype.name
            )
        self._variable = variable
        # 'delegate' means AutoCastVariable.op return self._variable.op, which
        # will raise an AttributeError in Eager (as intended). If set to any
        # other value, AutoCastVariable.op returns that value instead, which is
        # used to set the op attribute in AutoCastVariable.assign().
        self._op = "delegate"

    def _should_cast(self):
        """Returns True if this variable should be casted when accessed."""
        autocast_dtype = getattr(_autocast_dtype, "dtype", None)
        return autocast_dtype is not None and self.dtype != autocast_dtype

    @property
    def dtype(self):
        """The dtype of the underlying variable, before any casts are done."""
        return self._variable.dtype

    @property
    def true_dtype(self):
        """Deprecated alias of `dtype`."""
        return self._variable.dtype

    @property
    def _cast_dtype(self):
        dtype = getattr(_autocast_dtype, "dtype", None)
        return dtype or self._variable.dtype

    def value(self):
        val = self._variable.value()
        if not self._should_cast():
            return val
        return tf.cast(val, self._cast_dtype)

    def read_value(self):
        val = self._variable.read_value()
        return tf.cast(val, self._cast_dtype)

    def sparse_read(self, indices, name=None):
        """Reads the value of this variable sparsely, using `gather`."""
        val = self._variable.sparse_read(indices, name=name)
        return tf.cast(val, self._cast_dtype)

    def gather_nd(self, indices, name=None):
        """Gather slices of the variable into a Tensor."""
        val = self._variable.gather_nd(indices, name=name)
        return tf.cast(val, self._cast_dtype)

    def __getattr__(self, name):
        return getattr(self._variable, name)

    def _dense_var_to_tensor(self, dtype=None, name=None, as_ref=False):
        """Converts this variable to a tensor."""
        if as_ref:
            # This ValueError should not occur in practice since it is
            # impossible to pass as_ref=True using public APIs.
            raise ValueError(
                "Cannot convert AutoCastVariable to a tensor if "
                "as_ref=True is passed to convert_to_tensor"
            )
        if not self._should_cast():
            return tf.convert_to_tensor(self._variable, dtype=dtype, name=name)
        if dtype is not None and not dtype.is_compatible_with(self._cast_dtype):
            raise ValueError(
                "Incompatible type conversion requested to type {!r} for "
                "AutoCastVariable which is casted to type {!r}".format(
                    dtype.name, self._cast_dtype.name
                )
            )
        val = tf.convert_to_tensor(
            self._variable, dtype=self._variable.dtype, name=name
        )
        return tf.cast(val, self._cast_dtype)

    def __tf_tensor__(
        self,
        dtype: Optional[tf.dtypes.DType] = None,
        name: Optional[str] = None,
    ) -> tf.Tensor:
        return self._dense_var_to_tensor(dtype=dtype, name=name)

    def _should_act_as_resource_variable(self):
        """Pass resource_variable_ops.is_resource_variable check."""
        pass

    def __repr__(self):
        if tf.executing_eagerly() and not self._in_graph_mode:
            repr_str = (
                "<AutoCastVariable '{v.name}' shape={v.shape} "
                "dtype={v.dtype.name} dtype_to_cast_to={v._cast_dtype.name}, "
                "numpy={np_repr}>"
            )
            return repr_str.format(
                v=self, np_repr=numpy_text(self.read_value(), is_repr=True)
            )
        else:
            repr_str = (
                "<AutoCastVariable '{v.name}' shape={v.shape} "
                "dtype={v.dtype.name} dtype_to_cast_to={v._cast_dtype.name}>"
            )
            return repr_str.format(v=self)

    # Method delegations: We delegate the following methods to self._variable.
    # Each of these methods simply calls the same method on self._variable. The
    # base Variable raises NotImplementedError for most of these, so we must
    # override them.
    #
    # We do not define the following methods from Variable for the following
    # reasons:
    #   * 'count_up_to': This method only applies to int variables, which cannot
    #     be wrapped with an AutoCastVariable.
    #   * 'ref': Instead we inherit the definition from Variable.
    #     If we defined and delegated to Variable, the ref of an
    #     AutoCastVariable would be the same as the ref of the underlying
    #     variable, which would be strange as they are different Python objects.

    def set_shape(self, shape):
        return self._variable.set_shape(self, shape)

    @property
    def trainable(self):
        return self._variable.trainable

    @property
    def synchronization(self):
        return self._variable.synchronization

    @property
    def aggregation(self):
        return self._variable.aggregation

    def eval(self, session=None):
        return self._variable.eval(session)

    def initialized_value(self):
        return self._variable.initialized_value()

    @property
    def initial_value(self):
        return self._variable.initial_value

    @property
    def constraint(self):
        return self._variable.constraint

    def _apply_assign_update(
        self, update_fn, value, use_locking=None, name=None, read_value=True
    ):
        # TODO(b/146181571): This logic can be simplified once
        # DistributedVariable.assign returns a DistributedVariable. Currently
        # for MirroredStrategy, it returns a Mirrored value.
        if tf.compat.v1.executing_eagerly_outside_functions():
            assign_op = update_fn(value, use_locking, name, False)
            if read_value:
                # We create a new AutoCastVariable with the same underlying
                # tf.Variable.  The new AutoCastVariable is identical except the
                # 'op' attribute is defined. This matches the behavior of
                # tf.Variable.assign.
                var = create_autocast_variable(self._variable)
                var._op = assign_op
                return var
            return assign_op

        # Fallback to wrapping the returned variable in graph mode if possible
        assign_var = update_fn(value, use_locking, name, read_value)
        if read_value and tf.__internal__.ops.is_resource_variable(assign_var):
            return create_autocast_variable(assign_var)
        return assign_var

    def _apply_update(self, update_fn, *args, **kwargs):
        update_var = update_fn(*args, **kwargs)
        if tf.compat.v1.executing_eagerly_outside_functions():
            return self

        # Fallback to wrapping the returned variable in graph mode if possible
        if tf.__internal__.ops.is_resource_variable(update_var):
            return create_autocast_variable(update_var)
        return update_var

    def assign(self, value, use_locking=None, name=None, read_value=True):
        return self._apply_assign_update(
            self._variable.assign, value, use_locking, name, read_value
        )

    def assign_add(self, delta, use_locking=None, name=None, read_value=True):
        return self._apply_assign_update(
            self._variable.assign_add, delta, use_locking, name, read_value
        )

    def assign_sub(self, delta, use_locking=None, name=None, read_value=True):
        return self._apply_assign_update(
            self._variable.assign_sub, delta, use_locking, name, read_value
        )

    def scatter_sub(self, sparse_delta, use_locking=False, name=None):
        return self._apply_update(
            self._variable.scatter_sub, sparse_delta, use_locking, name
        )

    def scatter_add(self, sparse_delta, use_locking=False, name=None):
        return self._apply_update(
            self._variable.scatter_add, sparse_delta, use_locking, name
        )

    def scatter_max(self, sparse_delta, use_locking=False, name=None):
        return self._apply_update(
            self._variable.scatter_max, sparse_delta, use_locking, name
        )

    def scatter_min(self, sparse_delta, use_locking=False, name=None):
        return self._apply_update(
            self._variable.scatter_min, sparse_delta, use_locking, name
        )

    def scatter_mul(self, sparse_delta, use_locking=False, name=None):
        return self._apply_update(
            self._variable.scatter_mul, sparse_delta, use_locking, name
        )

    def scatter_div(self, sparse_delta, use_locking=False, name=None):
        return self._apply_update(
            self._variable.scatter_div, sparse_delta, use_locking, name
        )

    def scatter_update(self, sparse_delta, use_locking=False, name=None):
        return self._apply_update(
            self._variable.scatter_update, sparse_delta, use_locking, name
        )

    def batch_scatter_update(self, sparse_delta, use_locking=False, name=None):
        return self._apply_update(
            self._variable.batch_scatter_update, sparse_delta, use_locking, name
        )

    def scatter_nd_sub(self, indices, updates, name=None):
        return self._apply_update(
            self._variable.scatter_nd_sub, indices, updates, name
        )

    def scatter_nd_add(self, indices, updates, name=None):
        return self._apply_update(
            self._variable.scatter_nd_add, indices, updates, name
        )

    def scatter_nd_update(self, indices, updates, name=None):
        return self._apply_update(
            self._variable.scatter_nd_update, indices, updates, name
        )

    def load(self, value, session=None):
        return self._variable.load(value, session)

    @property
    def name(self):
        return self._variable.name

    @property
    def _shared_name(self):
        return self._variable._shared_name

    @property
    def initializer(self):
        return self._variable.initializer

    @property
    def device(self):
        return self._variable.device

    @property
    def op(self):
        if self._op == "delegate":
            return self._variable.op
        return self._op

    def _as_graph_element(self):
        graph_element = self._variable._as_graph_element()
        if graph_element is None:
            return self._op
        return graph_element

    @property
    def graph(self):
        return self._variable.graph

    @property
    def shape(self):
        return self._variable.shape

    def get_shape(self):
        return self._variable.get_shape()

    def __tf_tracing_type__(self, context):
        return AutoCastVariableSpec(self)

    def _gather_saveables_for_checkpoint(self):
        # By delegating this method to the wrapped variable, checkpoints with
        # AutoCastVariables are identical to checkpoints with normal variables.
        # Therefore models checkpointed with AutoCastVariables can be restored
        # on models with normal variables, and vice versa.
        return self._variable._gather_saveables_for_checkpoint()

    def _export_to_saved_model_graph(
        self, object_map, tensor_map, options, **kwargs
    ):
        # By delegating this method to the wrapped variable, SavedModel with
        # AutoCastVariables are identical to SavedModel with normal variables.
        resource_list = self._variable._export_to_saved_model_graph(
            object_map, tensor_map, options, **kwargs
        )
        object_map[self] = object_map[self._variable]
        return resource_list

    def _copy_trackable_to_cpu(self, object_map):
        """For implementing `Trackable`."""
        # Create a copy of `self._variable` to object_map, then create a new
        # copy of self that wraps the **copy** of `self._variable`.
        # When updating value, only the lowest-level variable will actually
        # update, since `AutoCastVariable` here is like a shell.
        self._variable._copy_trackable_to_cpu(
            object_map
        )  # pylint:disable=protected-access
        if self not in object_map:
            # If not populated already, populate self into the object map
            object_map[self] = AutoCastVariable(object_map[self._variable])

    # TODO(reedwm): Maybe encode the fact the variable is an AutoCastVariable in
    # to_proto().
    def to_proto(self, export_scope=None):
        return self._variable.to_proto(export_scope)

    def from_proto(self, variable_def, import_scope=None):
        return self._variable.from_proto(variable_def, import_scope)

    # Delegate the private attributes _handle_name and _initializer_op to
    # self._variable. SavedModel sets these attributes when loading a model. For
    # example, it sets _handle_name here:
    # https://github.com/tensorflow/tensorflow/blob/db26bd574fa95b5bdd53c08463dd19407cc0297e/tensorflow/python/tf_keras/saving/saved_model/load.py#L211
    # We need to expose these attributes on AutoCastVariable as well for
    # SavedModel to work properly.
    # TODO(reedwm/kathywu): Find a better way to support SavedModel. Exposing
    # private attributes is hacky and difficult to maintain.
    @property
    def _handle_name(self):
        return self._variable._handle_name

    @_handle_name.setter
    def _handle_name(self, handle_name):
        self._variable._handle_name = handle_name

    @property
    def _initializer_op(self):
        return self._variable._initializer_op

    @_initializer_op.setter
    def _initializer_op(self, initializer_op):
        self._variable._initializer_op = initializer_op

    # Operator overloads:
    # Note we only overload operators that support floating-point types, as
    # non-float variables cannot be wrapped with an AutoCastVariable.
    # Also note: We call read_value() instead of value(), because value() causes
    # gradients not to work properly when TPUStrategy is used: b/143380936

    def __add__(self, o):
        return self.read_value() + o

    def __radd__(self, o):
        return o + self.read_value()

    def __sub__(self, o):
        return self.read_value() - o

    def __rsub__(self, o):
        return o - self.read_value()

    def __mul__(self, o):
        return self.read_value() * o

    def __rmul__(self, o):
        return o * self.read_value()

    def __truediv__(self, o):
        return self.read_value() / o

    def __rtruediv__(self, o):
        return o / self.read_value()

    def __floordiv__(self, o):
        return self.read_value() // o

    def __rfloordiv__(self, o):
        return o // self.read_value()

    def __mod__(self, o):
        return self.read_value() % o

    def __rmod__(self, o):
        return o % self.read_value()

    def __lt__(self, o):
        return self.read_value() < o

    def __le__(self, o):
        return self.read_value() <= o

    def __gt__(self, o):
        return self.read_value() > o

    def __ge__(self, o):
        return self.read_value() >= o

    def __getitem__(self, o):
        return self.read_value()[o]

    def __pow__(self, o, modulo=None):
        return pow(self.read_value(), o, modulo)

    def __rpow__(self, o):
        return pow(o, self.read_value())

    def __neg__(self):
        return -self.read_value()

    def __abs__(self):
        return abs(self.read_value())

    def __div__(self, o):
        try:
            return self.read_value().__div__(o)
        except AttributeError:
            # See
            # https://docs.python.org/3/library/constants.html#NotImplemented
            return NotImplemented

    def __rdiv__(self, o):
        try:
            return self.read_value().__rdiv__(o)
        except AttributeError:
            # See
            # https://docs.python.org/3/library/constants.html#NotImplemented
            return NotImplemented

    def __matmul__(self, o):
        try:
            return self.read_value().__matmul__(o)
        except AttributeError:
            # See
            # https://docs.python.org/3/library/constants.html#NotImplemented
            return NotImplemented

    def __rmatmul__(self, o):
        try:
            return self.read_value().__rmatmul__(o)
        except AttributeError:
            # See
            # https://docs.python.org/3/library/constants.html#NotImplemented
            return NotImplemented


tf.register_tensor_conversion_function(
    AutoCastVariable, AutoCastVariable._dense_var_to_tensor
)


def create_autocast_variable(variable):
    """Creates an AutoCastVariable that wraps another variable.

    This typically just returns `AutoCastVariable(variable)`. But, if the
    variable is a DistributedVariable or one of its subclasses, we instead
    dynamically create a class that subclasses from both AutoCastVariable and
    variable.__class__. This is so the returned variable will still pass
    `isinstance(variable, variable.__class__)`, which is required for
    DistributedVariables and its subclasses to work properly.

    Args:
      variable: A floating-point resource variable to wrap.

    Returns:
      An AutoCastVariable that wraps the variable.
    """
    if not distributed_training_utils.is_distributed_variable(variable):
        return AutoCastVariable(variable)

    class AutoCastDistributedVariable(AutoCastVariable, variable.__class__):
        """An AutoCastVariable that also subclasses from variable.__class__.

        variable.__class__ is either a DistributedVariable or an
        AggregatingVariable.
        """

        def __repr__(self):
            return (
                "<AutoCastDistributedVariable dtype={v.dtype.name} "
                "dtype_to_cast_to={v._cast_dtype.name} "
                "inner_variable={v._variable}>"
            ).format(v=self)

    return AutoCastDistributedVariable(variable)


class enable_auto_cast_variables:
    """Context manager which enables the autocasting of `AutoCastVariable`s.

    Under this context manager, `AutoCastVariable`s will be cast to `dtype` if
    `dtype` is floating-point. Otherwise, `AutoCastVariable`s will not be cast.
    """

    __slots__ = ["_dtype", "_prev_dtype"]

    def __init__(self, dtype):
        if dtype and not dtype.is_floating:
            dtype = None
        self._dtype = dtype

    def __enter__(self):
        self._prev_dtype = getattr(_autocast_dtype, "dtype", None)
        _autocast_dtype.dtype = self._dtype

    def __exit__(self, type_arg, value_arg, traceback_arg):
        _autocast_dtype.dtype = self._prev_dtype




# Copyright 2018 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Part of TF-Keras training engine related to Python generators of array data.
"""

import functools
import math

import numpy as np
import tensorflow.compat.v2 as tf

from tf_keras.src import backend
from tf_keras.src import callbacks as cbks
from tf_keras.src.engine import training_utils
from tf_keras.src.engine import training_utils_v1
from tf_keras.src.utils import data_utils
from tf_keras.src.utils import generic_utils
from tf_keras.src.utils.mode_keys import ModeKeys

# isort: off
from tensorflow.python.platform import tf_logging as logging


def model_iteration(
    model,
    data,
    steps_per_epoch=None,
    epochs=1,
    verbose=1,
    callbacks=None,
    validation_data=None,
    validation_steps=None,
    validation_freq=1,
    class_weight=None,
    max_queue_size=10,
    workers=1,
    use_multiprocessing=False,
    shuffle=False,
    initial_epoch=0,
    mode=ModeKeys.TRAIN,
    batch_size=None,
    steps_name="steps",
    **kwargs,
):
    """Loop function for arrays of data with modes TRAIN/TEST/PREDICT.

    Args:
        model: TF-Keras Model instance.
        data: Either a tuple of NumPy/Tensor inputs (i.e. `(x,)` or `(x, y)` or
          `(x, y, sample_weights)`) or a generator or
          `keras.utils.data_utils.Sequence` object or Eager Iterator or Dataset.
        steps_per_epoch: Total number of steps (batches of samples) before
          declaring one epoch finished and starting the next epoch. Ignored with
          the default value of `None`.
        epochs: Number of times to iterate over the data.
        verbose: 0, 1, or 2. Verbosity mode.
          0 = silent, 1 = progress bar, 2 = one line per epoch.
          Note that the progress bar is not particularly useful when
          logged to a file, so verbose=2 is recommended when not running
          interactively (eg, in a production environment).
        callbacks: List of callbacks to be called during training.
        validation_data: Either a tuple of NumPy/Tensor inputs (i.e. `(x,)` or
          `(x, y)` or `(x, y, sample_weights)`) or a generator or
          `keras.utils.data_utils.Sequence` object or Eager Iterator or Dataset.
        validation_steps: Total number of steps (batches of samples) before
          declaring validation finished.
        validation_freq: Only relevant if validation data is provided. Integer
          or `collections.abc.Container` instance (e.g. list, tuple, etc.). If
          an integer, specifies how many training epochs to run before a new
          validation run is performed, e.g. `validation_freq=2` runs validation
          every 2 epochs. If a Container, specifies the epochs on which to run
          validation, e.g. `validation_freq=[1, 2, 10]` runs validation at the
          end of the 1st, 2nd, and 10th epochs.
        class_weight: Dictionary mapping class indices to a weight for the
            class.
        max_queue_size: Integer. Maximum size for the generator queue. If
          unspecified, `max_queue_size` will default to 10.
        workers: Integer. Maximum number of processes to spin up when using
          process-based threading. If unspecified, `workers` will default to 1.
          If 0, will execute the generator on the main thread.
        use_multiprocessing: Boolean. If `True`, use process-based threading. If
          unspecified, `use_multiprocessing` will default to `False`. Note that
          because this implementation relies on multiprocessing, you should not
          pass non-pickleable arguments to the generator as they can't be passed
          easily to children processes.
        shuffle: Boolean. Whether to shuffle the order of the batches at the
          beginning of each epoch. Only used with instances of `Sequence`
          (`keras.utils.Sequence`). Has no effect when `steps_per_epoch` is not
          `None`.
        initial_epoch: Epoch at which to start training (useful for resuming a
          previous training run).
        mode: One of ModeKeys.TRAIN/ModeKeys.TEST/ModeKeys.PREDICT.
        batch_size: Integer batch size or None if unknown. Will only be used if
          `data` is in NumPy/Tensor format.
        steps_name: The string name of the steps argument, either `steps`,
          `validation_steps`, or `steps_per_epoch`. Only used for error message
          formatting.
        **kwargs: Additional arguments for backwards compatibility. `steps` is
          accepted as an alias for `steps_per_epoch`.

    Returns:
        - In TRAIN mode: `History` object.
        - In TEST mode: Evaluation metrics.
        - In PREDICT mode: Outputs of the Model called on inputs.

    Raises:
        ValueError: in case of invalid arguments.
    """
    if "steps" in kwargs:
        steps_per_epoch = kwargs["steps"]

    # Determine the number of steps per epoch and whether we should reset the
    # dataset at the end of each epoch.
    reset_dataset_after_each_epoch = False
    original_dataset = None
    is_dataset = isinstance(data, (tf.data.Dataset, tf.compat.v1.data.Dataset))
    if is_dataset:
        original_dataset = data
        if steps_per_epoch is None:
            reset_dataset_after_each_epoch = True
            steps_per_epoch = training_utils_v1.infer_steps_for_dataset(
                model,
                data,
                steps_per_epoch,
                epochs=epochs,
                steps_name=steps_name,
            )

    # Convert to a format that supports `next(generator)`.
    generator, steps_per_epoch = convert_to_generator_like(
        data,
        steps_per_epoch=steps_per_epoch,
        batch_size=batch_size,
        epochs=epochs - initial_epoch,
        shuffle=shuffle,
    )

    do_validation = validation_data is not None
    is_sequence = isinstance(generator, data_utils.Sequence)
    _validate_arguments(
        is_sequence,
        is_dataset,
        use_multiprocessing,
        workers,
        steps_per_epoch,
        validation_data,
        validation_steps,
        mode,
        kwargs,
    )

    batch_function = _make_execution_function(
        model, mode, class_weight=class_weight
    )

    # Create the queue for the generator.
    enqueuer = None
    if not is_dataset:
        generator, enqueuer = _make_enqueued_generator(
            generator,
            workers=workers,
            use_multiprocessing=use_multiprocessing,
            max_queue_size=max_queue_size,
            shuffle=shuffle,
        )

    num_samples_or_steps, use_steps = _get_num_samples_or_steps(
        data, steps_per_epoch
    )

    count_mode = "steps" if use_steps else "samples"
    callbacks = cbks.configure_callbacks(
        callbacks,
        model,
        do_validation=do_validation,
        epochs=epochs,
        steps_per_epoch=steps_per_epoch,
        batch_size=batch_size,
        samples=num_samples_or_steps,
        count_mode=count_mode,
        verbose=verbose,
        mode=mode,
    )

    if mode == ModeKeys.PREDICT:
        aggregator = training_utils_v1.OutputsAggregator(
            True, steps=steps_per_epoch
        )
    else:
        aggregator = training_utils_v1.MetricsAggregator(
            True, steps=steps_per_epoch
        )

    should_set_learning_phase = tf.executing_eagerly() and model.run_eagerly
    if should_set_learning_phase:
        learning_phase_scope = backend.eager_learning_phase_scope(
            1 if mode == ModeKeys.TRAIN else 0
        )
        learning_phase_scope.__enter__()

    callbacks.model.stop_training = False
    callbacks._call_begin_hook(mode)

    initial_epoch = model._maybe_load_initial_epoch_from_ckpt(
        initial_epoch, mode
    )

    for epoch in range(initial_epoch, epochs):
        if callbacks.model.stop_training:
            break

        # Setup work for each epoch.
        model.reset_metrics()
        epoch_logs = {}
        if mode == ModeKeys.TRAIN:
            callbacks.on_epoch_begin(epoch, epoch_logs)

        if steps_per_epoch is None:
            # Loop over dataset until `OutOfRangeError` is raised.
            target_steps = np.inf
        else:
            # Loop over dataset for the specified number of steps.
            target_steps = steps_per_epoch

        step = 0
        while step < target_steps:
            batch_data = _get_next_batch(generator)
            if batch_data is None:
                if is_dataset:
                    # The dataset passed by the user ran out of batches.  Now we
                    # know the cardinality of the dataset.  If steps_per_epoch
                    # was specified, then running out of data is unexpected, so
                    # we stop training and inform the user.
                    if steps_per_epoch:
                        callbacks.model.stop_training = True
                        logging.warning(
                            "Your dataset ran out of data; interrupting "
                            "training. Make sure that your dataset can "
                            "generate at least `%s * epochs` batches (in "
                            "this case, %d batches). You may need to use "
                            "the repeat() function when building your dataset."
                            % (steps_name, steps_per_epoch * epochs)
                        )
                    elif step > 0:
                        steps_per_epoch = step
                        aggregator.steps = steps_per_epoch
                else:
                    # We ran out of batches while the user passed an iterator
                    # (legacy).
                    callbacks.model.stop_training = True
                    logging.warning(
                        "Your dataset iterator ran out of data; "
                        "interrupting training. Make sure that your iterator "
                        "can generate at least `%s * epochs` "
                        "batches (in this case, %d batches). You may need to"
                        "use the repeat() function when building your "
                        "dataset." % (steps_name, steps_per_epoch * epochs)
                    )
                break

            # `batch_size` used for validation data if validation
            # data is NumPy/EagerTensors.
            batch_size = int(tf.nest.flatten(batch_data)[0].shape[0])

            # Callbacks batch begin.
            batch_logs = {"batch": step, "size": batch_size}
            callbacks._call_batch_hook(mode, "begin", step, batch_logs)

            is_deferred = not model._is_compiled
            batch_outs = batch_function(*batch_data)
            if not isinstance(batch_outs, list):
                batch_outs = [batch_outs]

            if step == 0:
                aggregator.create(batch_outs)

                if is_deferred:
                    # Set callbacks params. We do this here when model is
                    # compiled only in the first iteration of this loop
                    # (deferred build scenario).
                    cbks.set_callback_parameters(
                        callbacks,
                        model,
                        do_validation=do_validation,
                        batch_size=batch_size,
                        epochs=epochs,
                        steps_per_epoch=steps_per_epoch,
                        samples=num_samples_or_steps,
                        verbose=verbose,
                        mode=mode,
                    )

            # Aggregate results.
            aggregator.aggregate(batch_outs)

            # Callbacks batch end.
            batch_logs = callbacks.make_logs(
                model, batch_logs, batch_outs, mode
            )
            callbacks._call_batch_hook(mode, "end", step, batch_logs)
            step += 1

            if callbacks.model.stop_training:
                break

        aggregator.finalize()
        results = aggregator.results
        epoch_logs = callbacks.make_logs(model, epoch_logs, results, mode)
        if len(results) == 1:
            results = results[0]

        # Run the test loop every epoch during training.
        if (
            do_validation
            and training_utils_v1.should_run_validation(validation_freq, epoch)
            and not callbacks.model.stop_training
        ):
            val_results = model_iteration(
                model,
                validation_data,
                steps_per_epoch=validation_steps,
                batch_size=batch_size,
                class_weight=class_weight,
                workers=workers,
                use_multiprocessing=use_multiprocessing,
                max_queue_size=max_queue_size,
                callbacks=callbacks,
                verbose=verbose,
                mode=ModeKeys.TEST,
                steps_name="validation_steps",
            )

            if not isinstance(val_results, list):
                val_results = [val_results]
            epoch_logs = callbacks.make_logs(
                model, epoch_logs, val_results, mode, prefix="val_"
            )

        if mode == ModeKeys.TRAIN:
            # Epochs only apply to `fit`.
            callbacks.on_epoch_end(epoch, epoch_logs)

        # Recreate dataset iterator for the next epoch.
        if reset_dataset_after_each_epoch and epoch < epochs - 1:
            generator = tf.compat.v1.data.make_one_shot_iterator(
                original_dataset
            )

    model._successful_loop_finish = True
    callbacks._call_end_hook(mode)

    if enqueuer is not None:
        enqueuer.stop()

    if should_set_learning_phase:
        learning_phase_scope.__exit__(None, None, None)

    if mode == ModeKeys.TRAIN:
        return model.history
    return results


# Maintain compatibility with the existing names.
fit_generator = functools.partial(model_iteration, mode=ModeKeys.TRAIN)
evaluate_generator = functools.partial(
    model_iteration, mode=ModeKeys.TEST, shuffle=False
)
predict_generator = functools.partial(
    model_iteration, mode=ModeKeys.PREDICT, shuffle=False
)


def _get_next_batch(generator):
    """Retrieves the next batch of input data."""
    try:
        generator_output = next(generator)
    except (StopIteration, tf.errors.OutOfRangeError):
        return None

    if not isinstance(generator_output, tuple):
        # Always wrap in a tuple.
        generator_output = (generator_output,)
    if len(generator_output) not in [1, 2, 3]:
        raise ValueError(
            "Output of generator should be a tuple of 1 or 2 or 3 "
            "elements: (input,) or (input, target) or "
            "(input, target, sample_weights). Received {}".format(
                generator_output
            )
        )
    return generator_output


def _validate_arguments(
    is_sequence,
    is_dataset,
    use_multiprocessing,
    workers,
    steps_per_epoch,
    validation_data,
    validation_steps,
    mode,
    kwargs,
):
    """Raises errors if arguments are invalid.

    Args:
      is_sequence: Boolean, whether data is a `keras.utils.data_utils.Sequence`
        instance.
      is_dataset: Boolean, whether data is a dataset instance.
      use_multiprocessing: Boolean. If `True`, use process-based threading. If
        unspecified, `use_multiprocessing` will default to `False`. Note that
        because this implementation relies on multiprocessing, you should not
        pass non-pickleable arguments to the generator as they can't be passed
        easily to children processes.
      workers: Integer. Maximum number of processes to spin up when using
        process-based threading. If unspecified, `workers` will default to 1. If
        0, will execute the generator on the main thread.
      steps_per_epoch: Total number of steps (batches of samples) before
        declaring one epoch finished and starting the next epoch. Ignored with
        the default value of `None`.
      validation_data: Either a tuple of NumPy/Tensor inputs (i.e. `(x,)` or
        `(x, y)` or `(x, y, sample_weights)`) or a generator or
        `keras.utils.data_utils.Sequence` object or Eager Iterator or Dataset.
      validation_steps: Total number of steps (batches of samples) before
        declaring validation finished.
      mode: One of ModeKeys.TRAIN/ModeKeys.TEST/ModeKeys.PREDICT.
      kwargs: Additional arguments for backwards compatibility.

    Raises:
      ValueError: If `steps_per_epoch` or `validation_steps` are not passed
        for data types that require them, or if unrecognized keyword
        arguments are passed.
    """
    if not is_sequence and use_multiprocessing and workers > 1:
        logging.warning(
            UserWarning(
                "Using a generator with `use_multiprocessing=True`"
                " and multiple workers may duplicate your data."
                " Please consider using the `keras.utils.Sequence`"
                " class."
            )
        )

    if steps_per_epoch is None and not is_dataset:
        arg_name = "steps_per_epoch" if mode == ModeKeys.TRAIN else "steps"
        raise ValueError(
            f"Please specify the number of steps via the `{arg_name}` argument."
        )

    val_gen = data_utils.is_generator_or_sequence(
        validation_data
    ) or isinstance(validation_data, tf.data.Iterator)
    if (
        val_gen
        and not isinstance(validation_data, data_utils.Sequence)
        and not validation_steps
    ):
        raise ValueError("Please specify the `validation_steps` argument.")

    if any(k != "steps" for k in kwargs):
        raise ValueError(
            f"Invalid arguments passed: {[k for k in kwargs if k != 'steps']}"
        )


def convert_to_generator_like(
    data, batch_size=None, steps_per_epoch=None, epochs=1, shuffle=False
):
    """Make a generator out of NumPy or EagerTensor inputs.

    Args:
      data: Either a generator or `keras.utils.data_utils.Sequence` object or
        `Dataset`, `Iterator`, or a {1,2,3}-tuple of NumPy arrays or
        EagerTensors.  If a tuple, the elements represent `(x, y,
        sample_weights)` and may be `None` or `[None]`.
      batch_size: Used when creating a generator out of tuples of NumPy arrays
        or EagerTensors.
      steps_per_epoch: Steps of the generator to run each epoch. If `None` the
        number of steps will be read from the data (for
        `keras.utils.data_utils.Sequence` types).
      epochs: Total number of epochs to run.
      shuffle: Whether the data should be shuffled.

    Returns:
      - Generator, `keras.utils.data_utils.Sequence`, or `Iterator`.

    Raises:
      - ValueError: If `batch_size` is not provided for NumPy or EagerTensor
        inputs.
    """
    if isinstance(data, tuple):
        # Scrub `Nones` that might have been passed for `targets`,
        # `sample_weights`.
        data = tuple(
            ele
            for ele in data
            if not all(e is None for e in tf.nest.flatten(ele))
        )

    if data_utils.is_generator_or_sequence(data) or isinstance(
        data, tf.data.Iterator
    ):
        if isinstance(data, data_utils.Sequence):
            if steps_per_epoch is None:
                steps_per_epoch = len(data)
        return data, steps_per_epoch
    if isinstance(data, tf.data.Dataset):
        return tf.compat.v1.data.make_one_shot_iterator(data), steps_per_epoch

    # Create generator from NumPy or EagerTensor Input.
    num_samples = int(tf.nest.flatten(data)[0].shape[0])
    if batch_size is None:
        raise ValueError(
            "When passing input data as arrays, do not specify "
            "`steps_per_epoch`/`steps` argument. "
            "Please use `batch_size` instead."
        )
    steps_per_epoch = int(math.ceil(num_samples / batch_size))

    def _gen(data):
        """Makes a generator out of a structure of NumPy/EagerTensors."""
        index_array = np.arange(num_samples)
        for _ in range(epochs):
            if shuffle:
                np.random.shuffle(index_array)
            batches = generic_utils.make_batches(num_samples, batch_size)
            for batch_start, batch_end in batches:
                batch_ids = index_array[batch_start:batch_end]
                flat_batch_data = training_utils.slice_arrays(
                    tf.nest.flatten(data), batch_ids, contiguous=(not shuffle)
                )
                yield tf.nest.pack_sequence_as(data, flat_batch_data)

    return _gen(data), steps_per_epoch


def _make_enqueued_generator(
    generator,
    workers=1,
    use_multiprocessing=False,
    max_queue_size=10,
    shuffle=False,
):
    """Create a buffered queue of next elements of the generator."""
    is_sequence = isinstance(generator, data_utils.Sequence)
    enqueuer = None
    if workers > 0:
        if is_sequence:
            enqueuer = data_utils.OrderedEnqueuer(
                generator,
                use_multiprocessing=use_multiprocessing,
                shuffle=shuffle,
            )
        else:
            enqueuer = data_utils.GeneratorEnqueuer(
                generator, use_multiprocessing=use_multiprocessing
            )
        enqueuer.start(workers=workers, max_queue_size=max_queue_size)
        output_generator = enqueuer.get()
    else:
        if is_sequence:
            output_generator = data_utils.iter_sequence_infinite(generator)
        else:
            output_generator = generator
    return output_generator, enqueuer


def _make_execution_function(model, mode, class_weight=None):
    """Makes function to run one step of model execution."""
    if mode == ModeKeys.TRAIN:
        f = functools.partial(model.train_on_batch, class_weight=class_weight)
    elif mode == ModeKeys.TEST:
        f = model.test_on_batch
    else:
        # Match signature of other modes to allow
        # 1, 2, or 3-tuples from generator
        def predict_on_batch(x, y=None, sample_weights=None):
            return model.predict_on_batch(x)

        f = predict_on_batch

    # Maintain stateful metrics across batch-level calls.
    if mode != ModeKeys.PREDICT:
        f = functools.partial(f, reset_metrics=False)

    return f


def _get_num_samples_or_steps(data, steps_per_epoch):
    """Returns number of samples or steps, and whether to use steps count
    mode."""
    flat_inputs = tf.nest.flatten(data)
    if hasattr(flat_inputs[0], "shape"):
        return int(flat_inputs[0].shape[0]), False
    return steps_per_epoch, True


class GeneratorOrSequenceTrainingLoop(training_utils_v1.TrainingLoop):
    """Generator-like.

    Input is Python generator, or Sequence object.

    The difference between this class and `GeneratorLikeTrainingFunction` is
    that this class only handles inputs that with x, y and sample_weight fused
    into one param.
    """

    def fit(
        self,
        model,
        x=None,
        y=None,
        batch_size=None,
        epochs=1,
        verbose=1,
        callbacks=None,
        validation_split=0.0,
        validation_data=None,
        shuffle=True,
        class_weight=None,
        sample_weight=None,
        initial_epoch=0,
        steps_per_epoch=None,
        validation_steps=None,
        validation_freq=1,
        max_queue_size=10,
        workers=1,
        use_multiprocessing=False,
    ):
        model._validate_or_infer_batch_size(batch_size, steps_per_epoch, x)
        training_utils_v1.check_generator_arguments(
            y, sample_weight, validation_split=validation_split
        )
        return fit_generator(
            model,
            x,
            steps_per_epoch=steps_per_epoch,
            epochs=epochs,
            verbose=verbose,
            callbacks=callbacks,
            validation_data=validation_data,
            validation_steps=validation_steps,
            validation_freq=validation_freq,
            class_weight=class_weight,
            max_queue_size=max_queue_size,
            workers=workers,
            use_multiprocessing=use_multiprocessing,
            shuffle=shuffle,
            initial_epoch=initial_epoch,
            steps_name="steps_per_epoch",
        )

    def evaluate(
        self,
        model,
        x=None,
        y=None,
        batch_size=None,
        verbose=1,
        sample_weight=None,
        steps=None,
        callbacks=None,
        max_queue_size=10,
        workers=1,
        use_multiprocessing=False,
    ):
        model._validate_or_infer_batch_size(batch_size, steps, x)
        training_utils_v1.check_generator_arguments(y, sample_weight)
        return evaluate_generator(
            model,
            x,
            steps=steps,
            verbose=verbose,
            callbacks=callbacks,
            max_queue_size=max_queue_size,
            workers=workers,
            use_multiprocessing=use_multiprocessing,
        )

    def predict(
        self,
        model,
        x,
        batch_size=None,
        verbose=0,
        steps=None,
        callbacks=None,
        max_queue_size=10,
        workers=1,
        use_multiprocessing=False,
    ):
        model._validate_or_infer_batch_size(batch_size, steps, x)
        return predict_generator(
            model,
            x,
            steps=steps,
            verbose=verbose,
            callbacks=callbacks,
            max_queue_size=max_queue_size,
            workers=workers,
            use_multiprocessing=use_multiprocessing,
        )


class EagerDatasetOrIteratorTrainingLoop(training_utils_v1.TrainingLoop):
    """A non-distributed Dataset or iterator in eager execution."""

    def fit(
        self,
        model,
        x=None,
        y=None,
        batch_size=None,
        epochs=1,
        verbose=1,
        callbacks=None,
        validation_split=0.0,
        validation_data=None,
        shuffle=True,
        class_weight=None,
        sample_weight=None,
        initial_epoch=0,
        steps_per_epoch=None,
        validation_steps=None,
        validation_freq=1,
        **kwargs,
    ):
        model._validate_or_infer_batch_size(batch_size, steps_per_epoch, x)
        # Make sure that y, sample_weights, validation_split are not passed.
        training_utils_v1.validate_dataset_input(
            x, y, sample_weight, validation_split
        )
        if (
            isinstance(x, (tf.compat.v1.data.Dataset, tf.data.Dataset))
            and shuffle
        ):
            training_utils_v1.verify_dataset_shuffled(x)

        return fit_generator(
            model,
            x,
            steps_per_epoch=steps_per_epoch,
            epochs=epochs,
            verbose=verbose,
            callbacks=callbacks,
            validation_data=validation_data,
            validation_steps=validation_steps,
            validation_freq=validation_freq,
            class_weight=class_weight,
            workers=0,
            shuffle=shuffle,
            initial_epoch=initial_epoch,
            steps_name="steps_per_epoch",
        )

    def evaluate(
        self,
        model,
        x=None,
        y=None,
        batch_size=None,
        verbose=1,
        sample_weight=None,
        steps=None,
        callbacks=None,
        **kwargs,
    ):
        model._validate_or_infer_batch_size(batch_size, steps, x)
        # Make sure that y, sample_weights, validation_split are not passed.
        training_utils_v1.validate_dataset_input(x, y, sample_weight)
        return evaluate_generator(
            model,
            x,
            steps=steps,
            verbose=verbose,
            workers=0,
            callbacks=callbacks,
        )

    def predict(
        self,
        model,
        x,
        batch_size=None,
        verbose=0,
        steps=None,
        callbacks=None,
        **kwargs,
    ):
        model._validate_or_infer_batch_size(batch_size, steps, x)
        return predict_generator(
            model,
            x,
            steps=steps,
            verbose=verbose,
            workers=0,
            callbacks=callbacks,
        )


class GeneratorLikeTrainingLoop(training_utils_v1.TrainingLoop):
    """TrainingLoop that handle inputs like python generator.

    This is the default handler for most of the input data types, includes
    symbolic tensors or Numpy array-like, Datasets and iterators in graph mode
    (since they generate symbolic tensors). This Function is used to handle
    model with `run_eagerly` = True.
    """

    def fit(
        self,
        model,
        x=None,
        y=None,
        batch_size=None,
        epochs=1,
        verbose=1,
        callbacks=None,
        validation_split=0.0,
        validation_data=None,
        shuffle=True,
        class_weight=None,
        sample_weight=None,
        initial_epoch=0,
        steps_per_epoch=None,
        validation_steps=None,
        validation_freq=1,
        **kwargs,
    ):
        batch_size = model._validate_or_infer_batch_size(
            batch_size, steps_per_epoch, x
        )
        x, y, sample_weights = model._standardize_user_data(
            x,
            y,
            sample_weight=sample_weight,
            class_weight=class_weight,
            batch_size=batch_size,
            check_steps=True,
            steps_name="steps_per_epoch",
            steps=steps_per_epoch,
            validation_split=validation_split,
            shuffle=shuffle,
        )

        if validation_data:
            validation_data = model._prepare_validation_data(
                validation_data, batch_size, validation_steps
            )
        elif validation_split and 0.0 < validation_split < 1.0:
            (
                x,
                y,
                sample_weights,
                val_x,
                val_y,
                val_sample_weights,
            ) = training_utils_v1.split_training_and_validation_data(
                x, y, sample_weights, validation_split
            )
            validation_data = (val_x, val_y, val_sample_weights)
        else:
            if validation_steps:
                raise ValueError(
                    "`validation_steps` should not be specified if "
                    "`validation_data` is None."
                )

        return fit_generator(
            model,
            (x, y, sample_weights),
            steps_per_epoch=steps_per_epoch,
            batch_size=batch_size,
            epochs=epochs,
            verbose=verbose,
            callbacks=callbacks,
            validation_data=validation_data,
            validation_steps=validation_steps,
            validation_freq=validation_freq,
            workers=0,
            shuffle=shuffle,
            initial_epoch=initial_epoch,
            steps_name="steps_per_epoch",
        )

    def evaluate(
        self,
        model,
        x=None,
        y=None,
        batch_size=None,
        verbose=1,
        sample_weight=None,
        steps=None,
        callbacks=None,
        **kwargs,
    ):
        batch_size = model._validate_or_infer_batch_size(batch_size, steps, x)
        x, y, sample_weights = model._standardize_user_data(
            x,
            y,
            sample_weight=sample_weight,
            batch_size=batch_size,
            check_steps=True,
            steps_name="steps",
            steps=steps,
        )
        return evaluate_generator(
            model,
            (x, y, sample_weights),
            steps=steps,
            batch_size=batch_size,
            verbose=verbose,
            workers=0,
            callbacks=callbacks,
        )

    def predict(
        self,
        model,
        x,
        batch_size=None,
        verbose=0,
        steps=None,
        callbacks=None,
        **kwargs,
    ):
        batch_size = model._validate_or_infer_batch_size(batch_size, steps, x)
        x, _, _ = model._standardize_user_data(
            x, check_steps=True, steps_name="steps", steps=steps
        )
        return predict_generator(
            model,
            x,
            steps=steps,
            batch_size=batch_size,
            verbose=verbose,
            workers=0,
            callbacks=callbacks,
        )




# Copyright 2018 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Part of the TF-Keras training engine related to distributed training."""

import numpy as np
import tensorflow.compat.v2 as tf

from tf_keras.src import backend
from tf_keras.src import callbacks as cbks
from tf_keras.src.distribute import distribute_coordinator_utils as dc
from tf_keras.src.distribute import distributed_training_utils_v1 as dist_utils
from tf_keras.src.engine import partial_batch_padding_handler as padding_util
from tf_keras.src.engine import training_arrays_v1
from tf_keras.src.engine import training_utils_v1
from tf_keras.src.utils.generic_utils import Progbar
from tf_keras.src.utils.mode_keys import ModeKeys

# isort: off
from tensorflow.python.distribute import input_lib
from tensorflow.python.platform import tf_logging as logging


def _per_replica_execution_function(model, mode):
    exec_func = model._make_execution_function(mode)
    return (
        exec_func.inputs,
        exec_func.outputs,
        exec_func.updates_op,
        exec_func.session_kwargs,
    )


def _build_model(strategy, model, mode, inputs, targets=None):
    if model._compile_distribution:
        dist_utils.clone_model_on_replicas(
            model, strategy, mode, inputs=inputs, targets=targets
        )
    else:
        dist_utils._build_distributed_network(
            model, strategy, mode, inputs, targets
        )


def _make_train_step_fn(model, mode, strategy, output_labels):
    """Create step fn.

    Args:
      model: a TF-Keras Model instance.
      mode: One of ModeKeys.TRAIN/ModeKeys.TEST/ModeKeys.PREDICT.
      strategy: a `tf.distribute.Strategy` instance.
      output_labels: the output labels for the step function.

    Returns:
      A step function to run by `tf.distribute.Strategy`.
    """

    def _step_fn(ctx, inputs):
        """A step fn that returns update ops."""
        if isinstance(inputs, (tuple, list)) and len(inputs) == 2:
            inputs, targets = inputs
        else:
            targets = None

        # When input feature is a dictionary of tensors, dictionary is
        # flattended to an array and passed as a model input. This results in
        # input mismatch when model input layer names are not sorted in
        # alphabetical order as `nest.flatten()`sorts dictionary elements by
        # keys. As so, transform input tensors into an array and order it along
        # `model._feed_input_names`.
        if isinstance(inputs, dict):
            inputs = [
                inputs[input_name] for input_name in model._feed_input_names
            ]

        _build_model(strategy, model, mode, inputs, targets)

        (
            grouped_inputs,
            grouped_outputs,
            grouped_updates,
            grouped_session_args,
        ) = strategy.extended.call_for_each_replica(
            _per_replica_execution_function,
            args=(dist_utils.get_distributed_model(model, mode), mode),
        )
        (
            all_inputs,
            all_outputs,
            all_updates,
            all_session_args,
        ) = dist_utils.unwrap_values(
            strategy,
            grouped_inputs,
            grouped_outputs,
            grouped_updates,
            grouped_session_args,
        )
        combined_fn = backend.function(
            all_inputs,
            all_outputs,
            updates=all_updates,
            name="distributed_" + str(mode) + "_function",
            **all_session_args
        )

        for label, output in zip(output_labels, combined_fn.outputs):
            if label == "loss":
                reduce_op = tf.distribute.ReduceOp.SUM
            else:
                # We reduce all other metrics using mean for now. This is
                # temporary workaround until new metrics are in place.
                reduce_op = tf.distribute.ReduceOp.MEAN
            ctx.set_last_step_output(label, output, reduce_op)

        # TODO(priyag, sourabhbajaj): Ignoring these things from the
        # combined_fn: feed_dict, session kwargs, run options, run_metadata for
        # now. These should be handled appropriately
        return combined_fn.updates_op

    return _step_fn


def experimental_tpu_fit_loop(
    model,
    dataset,
    epochs=100,
    verbose=1,
    callbacks=None,
    initial_epoch=0,
    steps_per_epoch=None,
    val_dataset=None,
    validation_steps=None,
    validation_freq=1,
):
    """Fit loop for training with TPU tf.distribute.Strategy.

    Args:
        model: TF-Keras Model instance.
        dataset: Dataset that returns inputs and targets
        epochs: Number of times to iterate over the data
        verbose: Integer, Verbosity mode, 0, 1 or 2
        callbacks: List of callbacks to be called during training
        initial_epoch: Epoch at which to start training
            (useful for resuming a previous training run)
        steps_per_epoch: Total number of steps (batches of samples)
            before declaring one epoch finished and starting the
            next epoch. Ignored with the default value of `None`.
        val_dataset: Dataset for validation data.
        validation_steps: Number of steps to run validation for
            (only if doing validation from data tensors).
            Ignored with the default value of `None`.
        validation_freq: Only relevant if validation data is provided. Integer
            or `collections.abc.Container` instance (e.g. list, tuple, etc.). If
            an integer, specifies how many training epochs to run before a new
            validation run is performed, e.g. `validation_freq=2` runs
            validation every 2 epochs. If a Container, specifies the epochs on
            which to run validation, e.g. `validation_freq=[1, 2, 10]` runs
            validation at the end of the 1st, 2nd, and 10th epochs.

    Returns:
        Returns `None`.

    Raises:
        ValueError: in case of invalid arguments.
    """
    mode = ModeKeys.TRAIN

    current_strategy = model._distribution_strategy
    iteration_value = min(
        steps_per_epoch, current_strategy.extended.steps_per_run
    )
    steps_per_run = backend.variable(
        value=iteration_value, dtype="int32", name="steps_per_run"
    )

    # TODO(fchollet): add support for `steps_per_epoch=None` in TPU loops.
    iterator = dist_utils.get_iterator(dataset, current_strategy)

    scope = dist_utils.distributed_scope(
        strategy=current_strategy, learning_phase=1
    )
    scope.__enter__()

    out_labels = model.metrics_names or []

    step_fn = _make_train_step_fn(
        model, ModeKeys.TRAIN, current_strategy, out_labels
    )

    # Add initial dummy values for loss and other metric tensors.
    initial_loop_values = {}
    initial_loop_values["loss"] = tf.constant(1e7)
    for m in model._get_training_eval_metrics():
        tensor = m.result()
        initial_loop_values[m.name] = tf.zeros(tensor.shape, tensor.dtype)

    ctx = current_strategy.extended.experimental_run_steps_on_iterator(
        step_fn,
        iterator,
        iterations=steps_per_run,
        initial_loop_values=initial_loop_values,
    )
    train_op = ctx.run_op
    output_tensors = ctx.last_step_outputs

    do_validation = bool(validation_steps)

    if model._compile_distribution:
        dist_utils._copy_weights_to_distributed_model(model, mode)

    callbacks = cbks.configure_callbacks(
        callbacks,
        model,
        do_validation=do_validation,
        epochs=epochs,
        steps_per_epoch=steps_per_epoch,
        verbose=verbose,
        count_mode="steps",
        mode=mode,
    )

    # Calculate the steps each time on the device.
    steps_to_run = [current_strategy.extended.steps_per_run] * (
        steps_per_epoch // current_strategy.extended.steps_per_run
    )
    if steps_per_epoch % current_strategy.extended.steps_per_run:
        steps_to_run.append(
            steps_per_epoch % current_strategy.extended.steps_per_run
        )
    target_steps = len(steps_to_run)

    callbacks._call_begin_hook(mode)

    initial_epoch = model._maybe_load_initial_epoch_from_ckpt(
        initial_epoch, mode
    )

    for epoch in range(initial_epoch, epochs):
        dist_utils._reset_metrics(model)
        callbacks.on_epoch_begin(epoch)
        epoch_logs = {}
        step_index = 0
        prev_step_count = None
        current_step = 0
        while current_step < target_steps:
            step_count = steps_to_run[current_step]
            batch_logs = {
                "batch": step_index,
                "size": 1,
                "num_steps": step_count,
            }
            callbacks._call_batch_hook(mode, "begin", step_index, batch_logs)
            if prev_step_count is None or step_count != prev_step_count:
                backend.get_session().run(steps_per_run.assign(step_count))
                prev_step_count = step_count
            try:
                _, outputs = backend.batch_get_value([train_op, output_tensors])
            except tf.errors.OutOfRangeError:
                logging.warning(
                    "Your dataset iterator ran out of data; "
                    "interrupting training. Make sure that your dataset "
                    "can generate at least `steps_per_epoch * epochs` "
                    "batches (in this case, %d batches)."
                    % steps_per_epoch
                    * epochs
                )
                break

            batch_logs.update(outputs)
            callbacks._call_batch_hook(mode, "end", step_index, batch_logs)
            step_index = step_index + step_count
            current_step += 1

            if callbacks.model.stop_training:
                break

        if do_validation and training_utils_v1.should_run_validation(
            validation_freq, epoch
        ):
            logging.info("Running validation at fit epoch: %s", epoch)

            if model._compile_distribution:
                # Since we create a new clone from the original model we need to
                # copy the weights back to the original model before we can run
                # validation.
                dist_utils._copy_weights_to_original_model(
                    model, ModeKeys.TRAIN
                )

            val_outs = experimental_tpu_test_loop(
                model,
                val_dataset,
                steps=validation_steps,
                verbose=verbose,
                callbacks=callbacks,
            )
            if not isinstance(val_outs, list):
                val_outs = [val_outs]
            # Same labels assumed.
            for label, val_out in zip(out_labels, val_outs):
                epoch_logs["val_" + label] = val_out

        callbacks.on_epoch_end(epoch, epoch_logs)
        if callbacks.model.stop_training:
            break
    model._successful_loop_finish = True
    callbacks._call_end_hook(mode)

    if model._compile_distribution:
        # Copy the weights back from the replicated model to the original model.
        dist_utils._copy_weights_to_original_model(model, ModeKeys.TRAIN)
    scope.__exit__(None, None, None)
    return model.history


def experimental_tpu_test_loop(
    model, dataset, verbose=0, steps=None, callbacks=None
):
    """Test loop for evaluating with TPU tf.distribute.Strategy.

    Args:
        model: TF-Keras Model instance.
        dataset: Dataset for input data.
        verbose: Integer, Verbosity mode 0 or 1.
        steps: Total number of steps (batches of samples)
            before declaring predictions finished.
            Ignored with the default value of `None`.
        callbacks: List of callbacks to be called during training

    Returns:
        Scalar loss (if the model has a single output and no metrics)
        or list of scalars (if the model has multiple outputs
        and/or metrics). The attribute `model.metrics_names` will give you
        the display labels for the outputs.
    """
    mode = ModeKeys.TEST
    current_strategy = model._distribution_strategy
    iterator = dist_utils.get_iterator(dataset, current_strategy)

    scope = dist_utils.distributed_scope(
        strategy=current_strategy, learning_phase=0
    )
    scope.__enter__()

    out_labels = model.metrics_names

    def _test_step_fn(inputs):
        """A fn that returns output of single test step."""
        if isinstance(inputs, (tuple, list)) and len(inputs) == 2:
            inputs, targets = inputs
        else:
            targets = None

        (
            tf.distribute.get_replica_context().merge_call(
                _build_model, args=(model, mode, inputs, targets)
            )
        )

        (_, outputs, updates, _) = _per_replica_execution_function(
            dist_utils.get_distributed_model(model, mode), mode
        )
        with tf.control_dependencies([updates]):
            return [tf.identity(out) for out in outputs]

    test_input_data = iterator.get_next()
    per_replica_outputs = current_strategy.run(
        _test_step_fn, args=(test_input_data,)
    )
    output_tensors = {}
    for label, output in zip(out_labels, per_replica_outputs):
        if label == "loss":
            reduce_op = tf.distribute.ReduceOp.SUM
        else:
            # We reduce all other metrics using mean for now. This is temporary
            # workaround until new metrics are in place.
            reduce_op = tf.distribute.ReduceOp.MEAN
        output_tensors[label] = current_strategy.reduce(
            reduce_op, output, axis=None
        )
    test_op = tf.group(list(output_tensors.values()))

    if verbose >= 1:
        progbar = Progbar(target=steps)

    if model._compile_distribution:
        dist_utils._copy_weights_to_distributed_model(model, mode)

    dist_utils._reset_metrics(model)

    callbacks = cbks.configure_callbacks(
        callbacks,
        model,
        do_validation=False,
        epochs=1,
        steps_per_epoch=steps,
        verbose=verbose,
        count_mode="steps",
        mode=ModeKeys.TEST,
    )
    callbacks._call_begin_hook(mode)

    outs = [0.0] * len(model.metrics_names)
    if steps is not None:
        target_steps = steps
    else:
        raise ValueError(
            "Number of steps could not be inferred from the data, "
            "please pass the steps argument."
        )

    current_step = 0
    while current_step < target_steps:
        batch_logs = {"batch": current_step, "size": 1}
        callbacks._call_batch_hook(mode, "begin", current_step, batch_logs)
        try:
            _, batch_outs = backend.batch_get_value([test_op, output_tensors])
        except tf.errors.OutOfRangeError:
            warning_msg = (
                "Make sure that your dataset can generate at least "
                "`steps` batches (in this case, {} batches).".format(steps)
            )

            logging.warning(
                "Your dataset iterator ran out of data; "
                "interrupting evaluation. " + warning_msg
            )
            target_steps = current_step
            break
        for i, label in enumerate(model.metrics_names):
            if i == 0:
                # Loss is stateless metrics.
                outs[i] += batch_outs[label]
            else:
                # For all stateful metrics, the aggregation is handled by
                # mirrored vars.
                outs[i] = batch_outs[label]

        batch_logs = callbacks.make_logs(model, batch_logs, outs, mode)
        callbacks._call_batch_hook(mode, "end", current_step, batch_logs)
        if verbose == 1:
            progbar.update(current_step + 1)
        current_step += 1

    if verbose >= 1:
        # Progress bar finishes at the end.
        progbar.update(target_steps)
    callbacks._call_end_hook(mode)

    scope.__exit__(None, None, None)
    if len(outs) > 0:
        outs[0] /= target_steps

    if len(outs) == 1:
        return outs[0]
    return outs


def experimental_tpu_predict_loop(
    model, dataset, verbose=0, steps=None, callbacks=None
):
    """Predict loop for predicting with TPU tf.distribute.Strategy.

    Args:
        model: TF-Keras Model instance.
        dataset: Dataset for input data.
        verbose: Integer, Verbosity mode 0 or 1.
        steps: Total number of steps (batches of samples)
            before declaring `_predict_loop` finished.
            Ignored with the default value of `None`.
        callbacks: List of callbacks to be called during training

    Returns:
        Array of predictions (if the model has a single output)
        or list of arrays of predictions
        (if the model has multiple outputs).
    """
    mode = ModeKeys.PREDICT
    dataset_fully_shaped = dist_utils.is_dataset_shape_fully_defined(dataset)
    padding_handler = None
    if not dataset_fully_shaped:
        # TODO(hongjunchoi): Investigate whether operations from
        # PartialBatchPaddingHandler are unnecessarily pruned out
        # during graph optimization.
        padding_handler = padding_util.PartialBatchPaddingHandler(
            model._feed_output_shapes
        )
        batch_size, _, prefetch_buffer = input_lib._get_dataset_attributes(
            dataset
        )
        padding_handler.padded_batch_size = batch_size
        padding_handler.padding_mask = dataset.reduce(
            padding_handler.padding_mask, padding_handler.update_mask
        )

        dataset = dataset.map(padding_handler.pad_batch)
        dataset = dataset.unbatch()
        # Upon this point, it is guaranteed that the dataset does not
        # have partial batches. Thus, we set `drop_remainder=True` to
        # get static shape information about the elements in the dataset.
        dataset = dataset.batch(batch_size, drop_remainder=True)

        if prefetch_buffer is not None:
            dataset = dataset.prefetch(prefetch_buffer)

    current_strategy = model._distribution_strategy
    iterator = dist_utils.get_iterator(dataset, current_strategy)

    scope = dist_utils.distributed_scope(
        strategy=current_strategy, learning_phase=0
    )
    scope.__enter__()

    def _predict_step_fn(inputs):
        """A fn that returns output of single prediction step."""

        (
            tf.distribute.get_replica_context().merge_call(
                _build_model, args=(model, mode, inputs)
            )
        )

        (_, outputs, updates, _) = _per_replica_execution_function(
            dist_utils.get_distributed_model(model, mode), mode
        )

        with tf.control_dependencies([updates]):
            return [tf.identity(out) for out in outputs]

    # TODO(hongjunchoi): When numpy array is passed as an input to `predict()`
    # use numpy arrays directly to avoid cumulating unnecessary input pipeline
    # ops.
    predict_input_data = iterator.get_next()
    per_replica_outputs = current_strategy.run(
        _predict_step_fn, args=(predict_input_data,)
    )
    output_tensors = dist_utils.flatten_per_replica_values(
        current_strategy, per_replica_outputs
    )

    if verbose >= 1:
        progbar = Progbar(target=steps)

    if model._compile_distribution:
        dist_utils._copy_weights_to_distributed_model(model, mode)

    dist_utils._reset_metrics(model)

    callbacks = cbks.configure_callbacks(
        callbacks,
        model,
        do_validation=False,
        epochs=1,
        steps_per_epoch=steps,
        verbose=verbose,
        count_mode="steps",
        mode=mode,
    )
    callbacks._call_begin_hook(mode)

    # Since we do not know how many samples we will see, we cannot pre-allocate
    # the returned Numpy arrays. Instead, we store one array per batch seen
    # and concatenate them upon returning.
    num_model_outputs = len(model.output_names)
    unconcatenated_outs = [[] for _ in range(num_model_outputs)]
    if steps is not None:
        target_steps = steps
    else:
        raise ValueError(
            "Number of steps could not be inferred from the data, "
            "please pass the steps argument."
        )

    current_step = 0
    while current_step < target_steps:
        batch_logs = {"batch": current_step, "size": 1}
        callbacks._call_batch_hook(mode, "begin", current_step, batch_logs)
        try:
            predict_ops = tf.group(output_tensors)
            _, batch_outs = backend.batch_get_value(
                [predict_ops, output_tensors]
            )

        except tf.errors.OutOfRangeError:
            warning_msg = (
                "Make sure that your dataset can generate at least "
                "`steps` batches (in this case, {} batches).".format(steps)
            )

            logging.warning(
                "Your dataset iterator ran out of data; "
                "interrupting evaluation. " + warning_msg
            )
            break

        # TODO(priyag): maybe need to unwrap the outputs first for
        # MirroredStrategy.
        for i in range(num_model_outputs):
            output_start_index = i * current_strategy.num_replicas_in_sync
            output_end_index = (
                output_start_index + current_strategy.num_replicas_in_sync
            )
            single_model_output = batch_outs[
                output_start_index:output_end_index
            ]
            unconcatenated_outs[i].extend(single_model_output)

        batch_logs = callbacks.make_logs(model, batch_logs, batch_outs, mode)
        callbacks._call_batch_hook(mode, "end", current_step, batch_logs)
        if verbose == 1:
            progbar.update(current_step + 1)
        current_step += 1

    if verbose >= 1:
        # Progress bar finishes at the end.
        progbar.update(current_step)

    callbacks._call_end_hook(mode)

    scope.__exit__(None, None, None)

    if len(unconcatenated_outs) == 1:
        prediction_result = np.concatenate(unconcatenated_outs[0], axis=0)
    else:
        prediction_result = [
            np.concatenate(out, axis=0) for out in unconcatenated_outs
        ]

    if padding_handler:
        prediction_result = padding_handler.apply_mask(prediction_result)

    return prediction_result


class DistributionSingleWorkerTrainingLoop(training_utils_v1.TrainingLoop):
    """Training loop for distribution strategy with single worker."""

    def fit(
        self,
        model,
        x=None,
        y=None,
        batch_size=None,
        epochs=1,
        verbose=1,
        callbacks=None,
        validation_split=0.0,
        validation_data=None,
        shuffle=True,
        class_weight=None,
        sample_weight=None,
        initial_epoch=0,
        steps_per_epoch=None,
        validation_steps=None,
        validation_freq=1,
        **kwargs
    ):
        """Fit loop for Distribution Strategies."""
        dist_utils.validate_callbacks(
            input_callbacks=callbacks, optimizer=model.optimizer
        )
        dist_utils.validate_inputs(x, y)

        batch_size, steps_per_epoch = dist_utils.process_batch_and_step_size(
            model._distribution_strategy,
            x,
            batch_size,
            steps_per_epoch,
            ModeKeys.TRAIN,
            validation_split=validation_split,
        )
        batch_size = model._validate_or_infer_batch_size(
            batch_size, steps_per_epoch, x
        )
        dataset = model._distribution_standardize_user_data(
            x,
            y,
            sample_weight=sample_weight,
            class_weight=class_weight,
            batch_size=batch_size,
            validation_split=validation_split,
            shuffle=shuffle,
            epochs=epochs,
        )
        if not dist_utils.is_distributing_by_cloning(model):
            with model._distribution_strategy.scope():
                (dataset, _, _) = model._standardize_user_data(
                    dataset,
                    sample_weight=sample_weight,
                    class_weight=class_weight,
                    batch_size=batch_size,
                    validation_split=validation_split,
                    shuffle=shuffle,
                )

        val_dataset = None
        if validation_data:
            (
                val_x,
                val_y,
                val_sample_weights,
            ) = training_utils_v1.unpack_validation_data(validation_data)
            dist_utils.validate_inputs(val_x, val_y)
            _, validation_steps = dist_utils.process_batch_and_step_size(
                model._distribution_strategy,
                val_x,
                batch_size,
                validation_steps,
                ModeKeys.TEST,
            )

            val_dataset = model._distribution_standardize_user_data(
                val_x,
                val_y,
                sample_weight=val_sample_weights,
                class_weight=None,
                batch_size=batch_size,
                validation_split=validation_split,
                shuffle=shuffle,
                allow_partial_batch=True,
            )
        elif validation_split:
            raise ValueError(
                "validation_split argument is not supported with "
                "distribution strategies."
            )

        if backend.is_tpu_strategy(model._distribution_strategy):
            steps_per_epoch = training_utils_v1.infer_steps_for_dataset(
                model,
                dataset,
                steps_per_epoch,
                epochs,
                steps_name="steps_per_epoch",
            )
            if steps_per_epoch is None:
                raise ValueError(
                    "Number of steps could not be inferred from the data, "
                    "please pass the steps_per_epoch argument."
                )

            if not tf.executing_eagerly():
                # Run TPU training in a custom loop in graph mode.
                return experimental_tpu_fit_loop(
                    model,
                    dataset,
                    epochs=epochs,
                    verbose=verbose,
                    callbacks=callbacks,
                    val_dataset=val_dataset,
                    initial_epoch=initial_epoch,
                    steps_per_epoch=steps_per_epoch,
                    validation_steps=validation_steps,
                    validation_freq=validation_freq,
                )

        return training_arrays_v1.fit_loop(
            model,
            dataset,
            batch_size=batch_size,
            epochs=epochs,
            verbose=verbose,
            callbacks=callbacks,
            val_inputs=val_dataset,
            shuffle=shuffle,
            initial_epoch=initial_epoch,
            steps_per_epoch=steps_per_epoch,
            validation_steps=validation_steps,
            validation_freq=validation_freq,
            steps_name="steps_per_epoch",
        )

    def evaluate(
        self,
        model,
        x=None,
        y=None,
        batch_size=None,
        verbose=1,
        sample_weight=None,
        steps=None,
        callbacks=None,
        **kwargs
    ):
        """Evaluate loop for Distribution Strategies."""
        dist_utils.validate_inputs(x, y)
        batch_size, steps = dist_utils.process_batch_and_step_size(
            model._distribution_strategy, x, batch_size, steps, ModeKeys.TEST
        )
        batch_size = model._validate_or_infer_batch_size(batch_size, steps, x)
        dataset = model._distribution_standardize_user_data(
            x,
            y,
            sample_weight=sample_weight,
            batch_size=batch_size,
            allow_partial_batch=True,
        )

        if backend.is_tpu_strategy(model._distribution_strategy):
            steps = training_utils_v1.infer_steps_for_dataset(
                model, dataset, steps, steps_name="steps"
            )
            if steps is None:
                raise ValueError(
                    "Number of steps could not be inferred from the data, "
                    "please pass the steps argument."
                )

            if not tf.executing_eagerly():
                # Run TPU evaluation in a custom loop in graph mode.
                return experimental_tpu_test_loop(
                    model,
                    dataset,
                    verbose=verbose,
                    steps=steps,
                    callbacks=callbacks,
                )

        return training_arrays_v1.test_loop(
            model,
            inputs=dataset,
            batch_size=batch_size,
            verbose=verbose,
            steps=steps,
            callbacks=callbacks,
        )

    def predict(
        self,
        model,
        x,
        batch_size=None,
        verbose=0,
        steps=None,
        callbacks=None,
        **kwargs
    ):
        """Predict loop for Distribution Strategies."""
        dist_utils.validate_inputs(x=x, y=None)
        batch_size, steps = dist_utils.process_batch_and_step_size(
            model._distribution_strategy, x, batch_size, steps, ModeKeys.PREDICT
        )
        batch_size = model._validate_or_infer_batch_size(batch_size, steps, x)
        dataset = model._distribution_standardize_user_data(
            x, batch_size=batch_size, allow_partial_batch=True
        )
        if backend.is_tpu_strategy(model._distribution_strategy):
            steps = training_utils_v1.infer_steps_for_dataset(
                model, dataset, steps, steps_name="steps"
            )
            if steps is None:
                raise ValueError(
                    "Number of steps could not be inferred from the data, "
                    "please pass the steps argument."
                )
            if not tf.executing_eagerly():
                return experimental_tpu_predict_loop(
                    model,
                    dataset,
                    verbose=verbose,
                    steps=steps,
                    callbacks=callbacks,
                )
        return training_arrays_v1.predict_loop(
            model,
            dataset,
            batch_size=batch_size,
            verbose=verbose,
            steps=steps,
            callbacks=callbacks,
        )


def _train_with_multi_worker(method):
    """Decorator handles multi worker training with distribution strategy."""

    def wrapper(model, **kwargs):
        def _worker_fn(_):
            callbacks = kwargs.pop("callbacks", None)
            filtered_callbacks = dist_utils.filter_distributed_callbacks(
                callbacks, model
            )
            kwargs["callbacks"] = filtered_callbacks
            return method(model, **kwargs)

        return dc.run_distribute_coordinator(
            _worker_fn, model._distribution_strategy
        )

    return wrapper


class DistributionMultiWorkerTrainingLoop(training_utils_v1.TrainingLoop):
    """Training loop for distribution strategy with multiple worker."""

    def __init__(self, single_worker_loop):
        self._single_worker_loop = single_worker_loop

    def fit(self, *args, **kwargs):
        return _train_with_multi_worker(self._single_worker_loop.fit)(
            *args, **kwargs
        )

    def evaluate(self, *args, **kwargs):
        return _train_with_multi_worker(self._single_worker_loop.evaluate)(
            *args, **kwargs
        )

    def predict(self, *args, **kwargs):
        # Currently predict is still using the single worker implementation.
        return self._single_worker_loop.predict(*args, **kwargs)

